/***************************************************************************
                         AC3_audio.h  -  description
                             -------------------
   Functions for handling AC3 sync frames.
                             -------------------
    begin                : Fri Aug 12 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _AC3_AUDIO_H_
#define _AC3_AUDIO_H_

#include "common/typedefs.h"
#include "MPEG2_SYS.h"

#define AC3_AUDIO_BSID        0x47
#define AC3_AUDIO_SYNC_BYTE1  0x0B
#define AC3_AUDIO_SYNC_BYTE2  0x77

#define AC3_AUDIO_BS_MS_CM    0x00
#define AC3_AUDIO_BS_MS_ME    0x01
#define AC3_AUDIO_BS_AS_VI    0x02
#define AC3_AUDIO_BS_AS_HI    0x03
#define AC3_AUDIO_BS_AS_D     0x04
#define AC3_AUDIO_BS_AS_C     0x05
#define AC3_AUDIO_BS_AS_E     0x06
#define AC3_AUDIO_BS_AS_VO    0x07
#define AC3_AUDIO_BS_MS_VO    0x07
#define AC3_AUDIO_BS(ptr_to_AC3_SYNC_FRAME) \
   ((ptr_to_AC3_SYNC_FRAME)->mode & 0x07)

#define AC3_AUDIO_AC_LFE      0x10
#define AC3_AUDIO_AC_1p1      0x00
#define AC3_AUDIO_AC_10       0x20
#define AC3_AUDIO_AC_20       0x40
#define AC3_AUDIO_AC_30       0x60
#define AC3_AUDIO_AC_21       0x80
#define AC3_AUDIO_AC_31       0xA0
#define AC3_AUDIO_AC_22       0xC0
#define AC3_AUDIO_AC_32       0xE0
#define AC3_AUDIO_AC(ptr_to_AC3_SYNC_FRAME) \
   ((ptr_to_AC3_SYNC_FRAME)->mode & 0xE0)

#define AC3_AUDIO_STATE_MASK           0xF000
#define AC3_AUDIO_STATE_INC            0x1000
#define AC3_AUDIO_next_frame_STATE     0x0000
#define AC3_AUDIO_SKIP_AUX_BYTES \
   (AC3_AUDIO_next_frame_STATE+6)
#define AC3_AUDIO_next_frame_END_STATE \
   (AC3_AUDIO_next_frame_STATE+AC3_AUDIO_STATE_INC)
#define AC3_AUDIO_read_BSI_STATE \
   (AC3_AUDIO_next_frame_STATE+AC3_AUDIO_STATE_INC)
#define AC3_AUDIO_read_BSI_END_STATE \
   (AC3_AUDIO_next_frame_STATE+(2*AC3_AUDIO_STATE_INC))
   
#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8  mode,
       num_ch,             /* Number of audio channels in the frame */
       bit_shift, data;

   u32 tmp_data,
       sample_rate,        /* in Hz */
       frame_len;          /* Total number of bytes in this frame */   
   u16 state,
       payload_len16;      /* Number of bytes in the payload */
   const
   u8 *payload;

   CLK_VAL  offset, pes_pts,
           *a_sync;
} AC3_SYNC_FRAME;


/* Find the first AC3 sync frame in the given data stream.
      INPUT: data_stream - Points to the data to look through.
             data_len    - Number of bytes in data_stream.
             frame       - Pointer to an AC3_SYNC_FRAME.
      RETURN: Success         - Returns number of bytes read from
                                  data_stream.
              Frame not found - 0 */
u32  AC3_AUDIO_next_frame(const u8       *data_stream,
                          u32             data_len,
                          AC3_SYNC_FRAME *frame);

/* Read the BSI data.
      INPUT: frame - Pointer to an AC3_SYNC_FRAME. */
void AC3_AUDIO_read_BSI(AC3_SYNC_FRAME *frame);

#ifdef __cplusplus
};
#endif

#endif

