/***************************************************************************
                          Thread.h  -  description
                             -------------------
   Provides a portable thread API.
                             -------------------
    begin                : Sat Jul 17 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __THREAD_H__
#define __THREAD_H__

#if defined(POSIX)
/************************Start POSIX Version************************/
# include <stdlib.h>
# include <pthread.h>

# define THREAD_FUNC_TYPE

/* Current thread will give up remaining processor time */
# define yield_THREAD()       pthread_yield()
/* Exits the current thread and returns the value. */
# define exit_THREAD(retval)  pthread_exit(retval)
/* Removes the thread from system. */
# define del_THREAD(t)        if (t) free(t)

/*************************End POSIX Version*************************/
#elif defined(WIN32)
/************************Start WIN32 Version************************/
# ifndef WIN32_LEAN_AND_MEAN
#  define WIN32_LEAN_AND_MEAN
# endif
# include <windows.h>
# include <process.h>

# define THREAD_FUNC_TYPE __stdcall

/* Current thread will give up remaining processor time */
# define yield_THREAD()       Sleep(0)
/* Exits the current thread and returns the value. */
# define exit_THREAD(retval)  _endthreadex((DWORD)retval)
/* Removes the thread from system. */
# define del_THREAD(t)        if (t) CloseHandle((HANDLE)t)

/*************************End WIN32 Version*************************/
#endif

/* The rest of this file is shared between all versions. */
#ifdef __cplusplus
extern "C" {
#endif

typedef void * THREAD;
typedef void *(THREAD_FUNC_TYPE *THREAD_FUNC)(void *);


/* Public Functions*/
/* Creates a new thread. */
THREAD new_THREAD(THREAD_FUNC f, void *args);
/* Current thread will wait for termination of given thread.
   Returns non-zero value for success. */
int    join_THREAD(THREAD t, void **retval);

/* The rest of these functions are implemented using the macros above. */
#ifdef IGNORE_THESE
/* Current thread will give up remaining processor time */
void   yield_THREAD();
/* Exits the current thread and returns the value. */
void   exit_THREAD(void *retval);
/* Removes the thread from system. */
void   del_THREAD(THREAD t);
#endif

#ifdef __cplusplus
};
#endif

#endif
