/***************************************************************************
                          ADAPT_FIELD.c  -  description
                             -------------------
   Stuff needed for handling ADAPTATION FIELDs.
                             -------------------
    begin                : Thu Jul 29 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <string.h>

#include "ADAPT_FIELD.h"


/* Reads the first ADAPTATION FIELD, in data_stream, into af.
      INPUT: data_stream - Points to the first byte of the ADAPTATION FIELD.
                             This function assumes there are at most
                             ADAPT_FIELD_MAX_SIZE bytes. This function also
                             assumes the first byte (adapt_field_len) is
                             non-zero.
      OUTPUT: af - Points to the ADAPT_FIELD that will hold data taken from
                     data_stream.
      RETURN: Success - A non-zero value
              Error   - 0 */
int ADAPT_FIELD_read(const u8 *data_stream, ADAPT_FIELD *af)
{
   const u8 *ptr;
   
   af->adapt_field_len = *data_stream;
   data_stream++;

   af->flags = *data_stream;
   if (!(af->flags & 0x1F))
      return 1;   /* ADAPTATION FIELD is empty */
      
   ptr = data_stream;   
   ptr++;

   if (ADAPT_FIELD_PCR_FLAG(af)) {
      af->PCR  = (s64)(*ptr) << 25;
      ptr++;

      af->PCR |= (s64)(*ptr) << 17;
      ptr++;

      af->PCR |= (s64)(*ptr) <<  9;
      ptr++;

      af->PCR |= (s64)(*ptr) <<  1;
      ptr++;

      af->PCR |= (s64)((*ptr >> 7) & 0x01);
      af->PCR *= 300;
      
      af->PCR += (s64)(*ptr & 0x01) << 8;
      ptr++;

      af->PCR += (s64)(*ptr);
      ptr++;
   }

   if (ADAPT_FIELD_OPCR_FLAG(af)) {
      af->OPCR  = (s64)(*ptr) << 25;
      ptr++;

      af->OPCR |= (s64)(*ptr) << 17;
      ptr++;

      af->OPCR |= (s64)(*ptr) <<  9;
      ptr++;

      af->OPCR |= (s64)(*ptr) <<  1;
      ptr++;

      af->OPCR |= (s64)((*ptr >> 7) & 0x01);
      af->OPCR *= 300;

      af->OPCR += (s64)(*ptr & 0x01) << 8;
      ptr++;

      af->OPCR += (s64)(*ptr);
      ptr++;
   }

   if (ADAPT_FIELD_SPLICING_POINT_FLAG(af)) {
      af->splice_countdown = (s8)(*ptr);
      ptr++;
   }

   if (ADAPT_FIELD_PRIV_DATA_FLAG(af)) {
      af->priv_data_len = *ptr;
      ptr++;

      if (af->priv_data_len > ADAPT_FIELD_MAX_PRIV_DATA_LEN)
         return 0;   /* too much private data */
      memcpy(af->priv_data, ptr, (size_t)af->priv_data_len);
      ptr += af->priv_data_len;
   }

   if (ADAPT_FIELD_EXT_FLAG(af)) {
      const u8 *ptr2;
      
      af->adapt_field_ext_len = *ptr;
      ptr++;

      ptr2          = ptr;
      af->ext_flags = *ptr;
      ptr++;

      if (ADAPT_FIELD_LTW_FLAG(af)) {
         af->ltw_valid_flag = (s8)(*ptr & 0x80);
         af->ltw_offset     = (u16)(*ptr & 0x7F) << 8;
         ptr++;
         
         af->ltw_offset    |= (u16)(*ptr);
         ptr++;
      }

      if (ADAPT_FIELD_PIECEWISE_RATE_FLAG(af)) {
         af->piecewise_rate  = (u32)(*ptr & 0x3F) << 16;
         ptr++;

         af->piecewise_rate |= (u32)(*ptr) << 8;
         ptr++;

         af->piecewise_rate |= (u32)(*ptr);
         ptr++;
      }

      if (ADAPT_FIELD_SEAMLESS_SPLICE_FLAG(af)) {
         af->splice_type  = (s8)((*ptr & 0xF0) >> 4);
         af->DTS_next_AU  = (s64)(*ptr & 0x0E) << 29;
         ptr++;

         af->DTS_next_AU |= (s64)(*ptr) << 22;
         ptr++;

         af->DTS_next_AU |= (s64)(*ptr & 0xFE) << 14;
         ptr++;

         af->DTS_next_AU |= (s64)(*ptr) << 7;
         ptr++;

         af->DTS_next_AU |= (s64)(*ptr >> 1);
         ptr++;
      }

      if ((ptr - ptr2) > af->adapt_field_ext_len)
         return 0;   /* more data than expected in the extension field */
   }

   /* make sure we did not read more data than expected */
   return ((ptr - data_stream) <= af->adapt_field_len);
}
