/***************************************************************************
                          PSI_SECTION.h  -  description
                             -------------------
   Stuff needed for handling PSI SECTIONs.
                             -------------------
    begin                : Sat Jul 31 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __PSI_SECTION_H__
#define __PSI_SECTION_H__

#include "common/typedefs.h"


#define PSI_SECTION_MAX_SIZE                 4096
#define PSI_SECTION_PROGRAM_ASSOCIATION_ID   0x00
#define PSI_SECTION_CONDITIONAL_ACCESS_ID    0x01
#define PSI_SECTION_PROGRAM_MAP_ID           0x02
#define PSI_SECTION_FORBIDDEN_ID             0xFF

#define PSI_SECTION_RES_ID(ptr_to_PSI_SECTION) \
   ( (0x02 < ptr_to_PSI_SECTION->table_id)&&(ptr_to_PSI_SECTION->table_id < 0x40) )
#define PSI_SECTION_PRIV_ID(ptr_to_PSI_SECTION) \
   ( (0x3F < ptr_to_PSI_SECTION->table_id)&&(ptr_to_PSI_SECTION->table_id < 0xFF) )

#define PSI_SECTION_SYNTAX_FLAG(ptr_to_PSI_SECTION) \
   (ptr_to_PSI_SECTION->flags & 0x80)
#define PSI_SECTION_PRIVATE_FLAG(ptr_to_PSI_SECTION) \
   (ptr_to_PSI_SECTION->flags & 0x40)
#define PSI_SECTION_VERSION_NUMBER(ptr_to_PSI_SECTION) \
   ((ptr_to_PSI_SECTION->flags >> 1) & 0x1F)
#define PSI_SECTION_CURRENT_NEXT_FLAG(ptr_to_PSI_SECTION) \
   (ptr_to_PSI_SECTION->flags & 0x01)

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
   u8  table_id,               /*   8: Identifies the content of this
                                         PSI_SECTION. */
       flags,
/*
      1   syntax_flag                  Set when this PSI_SECTION follows the
                                         generic section syntax.
      1   private_flag      <0> This value is used for all non-private sections
                            <X> In private sections, this bit is user defined
      5   version_number               Version number of the table this
                                         PSI_SECTION belongs to. When there is
                                         a change in the table, this value is
                                         incremented.
      1   current_next_flag            Set when this PSI_SECTION belongs to
                                         belongs to the currently active
                                         table. When not set, this PSI_SECTION
                                         belongs to the next table to become
                                         active.
*/
       section_number,          /*  8: Index of this PSI_SECTION, in the
                                         table. The first section is 0, the
                                         next is 1, and so on. */
       last_section_number;     /*  8: Index of the last PSI_SECTION in the
                                         table. A table with 10 sections will
                                         have a last_section_number value of
                                         9. */
                                         
   u16 payload_len16,           /* 12: Number of payload bytes.
                         <4093> Max value */
       table_id_extension;      /* 16: */
       
   const
   u8 *payload;                 /*     Points to the payload of this SECTION. */
} PSI_SECTION;


/* Reads the first PSI SECTION, in data_stream, into psi.
      INPUT: data_stream - Points to the first byte of the PSI SECTION.
                             This function assumes there are at most
                             8 bytes.
      OUTPUT: psi - Points to the PSI_SECTION that will hold data taken from
                      data_stream.
      RETURN: Success - A non-zero value
              Error   - 0 */
int PSI_SECTION_read(const u8 *data_stream, PSI_SECTION *psi);

/* Checks the CRC value of the given psi.
      INPUT: psi - Points to the PSI_SECTION that needs to be checked.
      RETURN: CRC good - A non-zero value
              CRC bad  - 0 */
int PSI_SECTION_chk_crc(PSI_SECTION *psi);

#ifdef __cplusplus
};
#endif

#endif
