/***************************************************************************
                          PSI_module.h  -  description
                             -------------------
   PSI SYSTEM module functions.
                             -------------------
    begin                : Sat Aug 14 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __PSI_MODULE_H__
#define __PSI_MODULE_H__

#include "common/typedefs.h"
#include "MPEG2_SYS.h"


/* Use this macro to initialize a MPEG2_SYS_MODULE structure. */
#define PSI_SYS_MODULE \
{\
   PSI_SYS_MODULE_psi_pes_module, \
   PSI_SYS_MODULE_process_pkt, \
   {\
      PSI_SYS_MODULE_mem_needed, \
      PSI_SYS_MODULE_init, \
      PSI_SYS_MODULE_done, \
      NULL, \
      PSI_SYS_MODULE_discontinuity, \
      NULL \
   }\
}

#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t PSI_SYS_MODULE_mem_needed();

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void   PSI_SYS_MODULE_init(void *mem, SYS_DATA *sys_data);

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void   PSI_SYS_MODULE_done(void *mem);

/* Alerts the module to a discontinuity in the data stream.
      INPUT: void * - Pointer to the memory provided for the module. */
void   PSI_SYS_MODULE_discontinuity(void *mem);


/**************************************************************************
   SYSTEM MODULE FUNCTIONS
     These functions are used by the MPEG2 TS decoder to extract the PSI
       data from the corresponding TS_PKTs, and create PSI_SECTIONs for
       the PSI mdules.
 **************************************************************************/
/* Provides the PSI module that will handle the PSI_SECTIONs.
      INPUT: void *                 - Pointer to the memory provided for
                                        the module.
             MPEG2_PSI_PES_MODULE * - Pointer to a PSI module. */
void PSI_SYS_MODULE_psi_pes_module(void *mem, MPEG2_PSI_PES_MODULE *psi);

/* Processes the packet handled by the module.
      INPUT: void *   - Pointer to the memory provided for the module.
             TS_PKT * - Pointer to packet that needs handling. */
void PSI_SYS_MODULE_process_pkt(void *mem, TS_PKT *ts);

#ifdef __cplusplus
};
#endif

#endif
