/***************************************************************************
                          mpeg_ps.h  -  description
                             -------------------
   Functions for handling mpeg packet streams.
                             -------------------
    begin                : Fri Mar 22 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef _MPEG_PS_H_
#define _MPEG_PS_H_

#include "common/typedefs.h"
#include "common/Thread.h"
#include "common/File.h"
#include "PES_PKT.h"


#ifdef __cplusplus
extern "C" {
#endif

/* A player function to be used with the MPEG2_SYS_Play() function.
     This function can handle MPEG packet streams. */
void * THREAD_FUNC_TYPE MPEG_PS_Play(void *arg);

/* Creates a program, using the first audio and video streams
     found in the given MPEG packet stream.
      INPUT: fp - Points to the FILE with the MPEG packet stream.
      RETURN: Program number, of newly created program, is stored
                in the global struct (sys.program_number).
              Success - A non-zero value
              Failure - 0 */
void MPEG_PS_Create_Program(FILE_PTR fp);

/* Gets the stream ids and types from the packet stream.
      INPUT: fp   - Points to the FILE with the MPEG packet
                      stream.
             size - Number of elements in the PID and stream_type
                      arrays.
      OUTPUT: PID         - Array of stream ids found in the packet
                              stream.
              stream_type - Array of STREAM_TYPE_* values 
                              corresponding to the values in the
                              PID array.
      RETURN: The number of PIDs found. */
u32  MPEG_PS_get_PIDs(FILE_PTR fp, u16 *PID, u8 *stream_type, u32 size);

/* Tries to determine the type of data contained in the stream with
     the given PID.
      INPUT: fp  - Pointer to an open file that contains MPEG PS data.
             PID - PID of the stream to look for.
      RETURN: Success - One of the STREAM_TYPE_* values defined
                          in PMT_SECTION.h
              Failure - Zero. */
u8   MPEG_PS_stream_type(FILE_PTR fp, u16 PID);


/* Deals with the PES_PKTs decoded in MPEG_PS_decode().
      INPUT PES_PKT * - Pointer to the latest PES_PKT in the stream.
            void *    - Pointer to any private data needed by the
                          function.
      RETURN: Non-zero - Ready for the next PES_PKT.
              Zero     - Stop decoding PES_PKTs. */
typedef int (*MPEG_PS_decode_func)(PES_PKT *, void *);

/* Decodes the MPEG program stream in the given file
      INPUT: fp      - Pointer to an open file that contains MPEG PS 
                         data.
             end_pos - Point at which the decoding will stop. Decoding
                         will stop at an earlier point, if an EOF is
                         encountered.
             func    - Function that will deal with the decoded 
                         PES_PKTs.
             arg     - Pointer to any private data for func. */
void MPEG_PS_decode(FILE_PTR            fp, 
                    FILE_POS            end_pos, 
                    MPEG_PS_decode_func func,
                    void               *arg);

#ifdef __cplusplus
};
#endif

#endif

