/***************************************************************************
                          output_blocks.c  -  description
                             -------------------
   Functions used for placing decoded blocks into the frame buffer.
                             -------------------
    begin                : Tue Dec 14 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifdef _M1V_TRACE_
#include <stdio.h>
#endif
#include <stdlib.h>

#include "output_blocks.h"
#include "mpeg2_video/motion_vectors.h"


#define HALF_PIXEL_X             0x01
#define HALF_PIXEL_Y             0x02
#define HALF_PIXEL_MASK          0x03
#define HALF_PIXEL_COLOUR_SHIFT     2


/* Place a decoded Block into the frame buffer.
      INPUT: slice - Points to the SLICE holding the decoded Block,
                       and the pointer to the frame buffer. */
void output_block1(SLICE1 *slice)
{
   s16 *in0;
   u8  *out0;
   int  wid, dct_type;

   /* do a boundary check, before writing to frame buffer */
   if (slice->cur_row >= slice->mb_hgt)
      return;

   /* initialize some stuff */
   out0 = slice->pred_blk;
   wid  = (int)slice->frame_wid;
   if (slice->cc)
      wid >>= 1;

#ifdef _M1V_TRACE_
   printf("BLK: pred_blk = %u  out0 = %u\n",
          (u32)slice->pred_blk, (u32)out0);
#endif

   if (!MB_INTRA(slice)) {    /* do the motion vector stuff */
      int move_blk;
      s16 pred_blk[64], *in1;
      u16 idx;
      u8  shift, move_type;
      
      in0 = pred_blk;
      if (!slice->i) {
         s16 mv_x, mv_y,
             min_x = (s16)slice->cur_col * -16,
             min_y = (s16)slice->cur_row * -16,
             max_x = (s16)(slice->seq_hdr->horiz_size - 16) + min_x,
             max_y = (s16)(slice->seq_hdr->vert_size  - 16) + min_y;
             
#ifdef _M1V_TRACE_
         printf("MV_WID: wid=%d\n", wid);
         printf("MV_LIMITS: min_x=%hd max_x=%hd  min_y=%hd max_y=%hd\n",
                min_x, max_x, min_y, max_y);
#endif

         /* use vector[]s to move mv_blk[]s to new locations */
         for (idx = 0; idx < 2; idx++) {
            if (slice->mv_blk[idx]) {
               slice->move[idx] = 0;
               shift            = 0;
               
               mv_x = slice->vector[idx][0] >> 1;
               mv_y = slice->vector[idx][1] >> 1;

               if (slice->vector[idx][0] & 0x0001) {
                  shift |= HALF_PIXEL_X;
                  if (mv_x > (max_x-1)) {
                     mv_x = max_x-1;
                     goto MV_X_IS_GOOD;
                  }
               }
               else if (mv_x > max_x) {
                  mv_x = max_x;
                  goto MV_X_IS_GOOD;
               }
               if (mv_x < min_x)
                  mv_x = min_x;
MV_X_IS_GOOD:
               if (slice->vector[idx][1] & 0x0001) {
                  shift |= HALF_PIXEL_Y;
                  if (mv_y > (max_y-1)) {
                     mv_y = max_y-1;
                     goto MV_Y_IS_GOOD;
                  }
               }
               else if (mv_y > max_y) {
                  mv_y = max_y;
                  goto MV_Y_IS_GOOD;
               }
               if (mv_y < min_y)
                  mv_y = min_y;
MV_Y_IS_GOOD:
#ifdef _M1V_TRACE_
               printf("MV: horiz = %hd  vert = %hd  idx=%hd\n",
                      slice->vector[idx][0] >> 1,
                      slice->vector[idx][1] >> 1, idx);
               printf("          = %hd       = %hd\n", mv_x, mv_y);
               printf("MV: mv_blk = %u\n", (u32)slice->mv_blk[idx]);
#endif
               if ( (mv_x)||(mv_y) ) {
                  move_blk  = (int)mv_x + ((int)mv_y * (int)wid);
                  slice->mv_blk[idx] += move_blk;

                  /* change from (DIV 2) to (/ 2) */
                  if ( (mv_x < 0)&&(shift & HALF_PIXEL_X) )
                     mv_x++;
                  if ( (mv_y < 0)&&(shift & HALF_PIXEL_Y) )
                     mv_y++;
                  
                  move_blk -= (int) mv_x >> 1;
                  move_blk -= (int)(mv_y >> 1) * (int)(wid >> 1);
                  slice->move[idx]   += move_blk;
#ifdef _M1V_TRACE_
                  printf("           = %u\n", (u32)slice->mv_blk[idx]);
#endif
                  if (mv_x & 0x0001)
                     shift |= HALF_PIXEL_X << HALF_PIXEL_COLOUR_SHIFT;
                  if (mv_y & 0x0001)
                     shift |= HALF_PIXEL_Y << HALF_PIXEL_COLOUR_SHIFT;
               }
               slice->half_pixel[idx] = shift;
            }
         }
      }

      idx = 0;
      switch (slice->i) {
      case 0:
         move_blk  =  8;
         break;
      case 1:
         move_blk  = wid << 3;
         move_blk -=  8;
         break;
      case 2:
         move_blk  =  8;
         break;
      case 3:
         move_blk  = wid;
         move_blk *=(12 * (int)slice->cur_row) + 8;
         move_blk += (1 + (int)slice->cur_col) << 3;
         move_blk  = (int)((slice->luma_size * 5) >> 2) - move_blk;
         break;
      case 4:
         move_blk  = (int)slice->luma_size / -4;
         
         slice->half_pixel[0] >>= HALF_PIXEL_COLOUR_SHIFT;
         slice->half_pixel[1] >>= HALF_PIXEL_COLOUR_SHIFT;
         break;
      case 5:
         move_blk  =  0;
      }
      shift = ( (slice->mv_blk[0])&&(slice->mv_blk[1]) ) ? 1 : 0;

#ifdef _M1V_TRACE_
      printf("MV: mv_blk0=%u mv_blk1=%u\n",
             (u32)slice->mv_blk[0], (u32)slice->mv_blk[1]);
#endif

      /* get the predictor block */
      MV_CLEAR_BLOCK(in0);
      do {
         if (slice->mv_blk[idx]) {
            move_type = slice->half_pixel[idx] & HALF_PIXEL_MASK;
#ifdef _M1V_TRACE_
            printf("    half%hu=%hu\n", idx, (u16)move_type);

            if (slice->mv_blk[idx] < (u8 *)10000000)
               printf("MV_ERROR: mv_blk[%d]=%u < 1buf __%hu\n", idx, (u32)slice->mv_blk[idx], (u16)slice->half_pixel[idx]);
            else if (slice->mv_blk[idx] < (u8 *)20000000) {
               if ((slice->mv_blk[idx]+(wid*7)+8) > (u8 *)(10000000+slice->frame_size))
                  printf("MV_ERROR: mv_blk[%d]=%u >= 1buf+%u __%hu\n", idx, (u32)slice->mv_blk[idx], slice->frame_size, (u16)slice->half_pixel[idx]);
            }
            else if ((slice->mv_blk[idx]+(wid*7)+8) > (u8 *)(20000000+slice->frame_size))
               printf("MV_ERROR: mv_blk[%d]=%u >= 2buf+%u __%hu\n", idx, (u32)slice->mv_blk[idx], slice->frame_size, (u16)slice->half_pixel[idx]);
#else
            if (!move_type) {
               MV_PRED___(in0, slice->mv_blk[idx], wid);
            }
            else if (move_type == HALF_PIXEL_X) {
               MV_PRED_X_(in0, slice->mv_blk[idx], wid);
            }
            else if (move_type == HALF_PIXEL_Y) {
               MV_PRED__Y(in0, slice->mv_blk[idx],wid, wid);
            }
            else {
               MV_PRED_XY(in0, slice->mv_blk[idx], wid, wid);
            }
#endif
            if (slice->i == 3)
               slice->mv_blk[idx] -= slice->move[idx];
            slice->mv_blk[idx] += move_blk;
         }
         idx++;
      } while (idx&1);

#ifdef _M1V_TRACE_
      printf("           =%u        =%u\n",
             (u32)slice->mv_blk[0], (u32)slice->mv_blk[1]);
#endif
      in0 = slice->QFS;
      in1 = pred_blk;
#ifdef _M1V_TRACE_
      if (out0 < slice->pred_frame)
         printf("SAT_ERROR: out0 < pred_frame  wid=%d\n", wid);
      else if ((out0+(wid*7)+8) > (slice->pred_frame+slice->frame_size))
         printf("SAT_ERROR: out0 > pred_frame+%u  wid=%d\n", slice->frame_size, wid);
#else
      /* put the final block into the frame */
      MV_ADD_PRED(in0, in1, out0, wid, shift);
#endif
   }
   else {
      in0 = slice->QFS;
#ifdef _M1V_TRACE_
      if (out0 < slice->pred_frame)
         printf("SAT_ERROR: out0 < pred_frame  wid=%d\n", wid);
      else if ((out0+(wid*7)+8) > (slice->pred_frame+slice->frame_size))
         printf("SAT_ERROR: out0 > pred_frame+%u  wid=%d\n", slice->frame_size, wid);
#else
      /* put the final block into the frame */
      MV_SATURATION(in0, out0, wid);
#endif
   }

   /* move frame pointers into correct position */
   switch (slice->i) {
   case 0:
      wid   =  8;
      break;
   case 1:
      wid <<=  3;
      wid  -=  8;
      break;
   case 2:
      wid   =  8;
      break;
   case 3:
      wid  *=(12 * (int)slice->cur_row) + 8;
      wid  += (1 + (int)slice->cur_col) << 3;
      wid   = (int)((slice->luma_size * 5) >> 2) - wid;
      break;
   case 4:
      wid   = (int)slice->luma_size / -4;
      break;
   case 5:
      slice->ref_blkF += 16;
      slice->ref_blkB += 16;
      if ((slice->cur_col+1) == slice->mb_wid) {
         dct_type = 30 * wid;
         wid *=(24 * (int)slice->cur_row) + 30;

         dct_type += (int)slice->frame_wid - (int)slice->seq_hdr->horiz_size;
         wid      += (int)slice->frame_wid - (int)slice->seq_hdr->horiz_size;
         slice->ref_blkF += dct_type;
         slice->ref_blkB += dct_type;
      }
      else {
         dct_type = 0;
         wid *= 24 * (int)slice->cur_row;
      }
      wid  += (2 + (int)slice->cur_col) << 3;
      wid  -= (int)slice->luma_size;

      if (dct_type) {
         slice->cur_col = 0;
         slice->cur_row++;
      }
      else
         slice->cur_col++;
#ifdef _M1V_TRACE_
      printf("MV_REF: blkF=%u  blkB=%u\n", (u32)slice->ref_blkF, (u32)slice->ref_blkB);
#endif
   }

   slice->pred_blk += wid;
#ifdef _M1V_TRACE_
   printf("              = %u\n", (u32)slice->pred_blk);
#endif
}

/* Place skipped Macro-Blocks into the frame buffer.
      INPUT: slice - Points to the SLICE holding the pointer to the
                       frame buffer.
             cnt   - Number of Macro-Blocks that were skipped. */
void skip_macroblocks1(SLICE1 *slice, u32 cnt)
{
   u32 inc = 0;
   u8  mb_flags;

   MV_CLEAR_BLOCK(slice->QFS);

   if (PIC_HEADER1_CODING_TYPE(slice->pic_hdr) == CODING_TYPE_B_FRAME)
      mb_flags = slice->mb_flags;
   else {
      slice->vector[0][0] = 0;
      slice->vector[0][1] = 0;
      mb_flags = MB_MOTION_FWD_BIT;
   }
   
   slice->mb_flags  = 0;
   slice->mv_blk[0] = NULL;
   slice->mv_blk[1] = NULL;
   do {
      if (mb_flags & MB_MOTION_FWD_BIT)
         slice->mv_blk[0] = slice->ref_blkF + inc;
      if (mb_flags & MB_MOTION_BCK_BIT)
         slice->mv_blk[1] = slice->ref_blkB + inc;

#ifdef _M1V_TRACE_
      printf("SKIPPED MACROBLOCK\n"
             "  cur_col= %hu, cur_row= %hu\n",
             slice->cur_col, slice->cur_row);
#endif
      for (slice->i = 0; slice->i < 6; slice->i++) {
         if (slice->i < 4)
            slice->cc = 0;
         else if (slice->i & 0x01)
            slice->cc = 2;
         else
            slice->cc = 1;
# ifdef _M1V_TRACE_
         printf("SKIPPED BLOCK %hu\n", (u16)slice->i);
# endif
         output_block1(slice);
      }
      cnt--;
   } while (cnt);
}

#undef HALF_PIXEL_X
#undef HALF_PIXEL_Y
#undef HALF_PIXEL_COLOUR_MASK
#undef HALF_PIXEL_COLOUR_SHIFT

