/***************************************************************************
                          motion_vectors.h  -  description
                             -------------------
   These are the routines needed to handle motion vectors.
                             -------------------
    begin                : Wed Nov 24 2004
    copyright            : (C) 2004 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __MOTION_VECTORS_H__
#define __MOTION_VECTORS_H__

#include "common/typedefs.h"

#if defined(__MMX__)
# if defined(__GCC__)
#  define MV_DONE    __asm__ __volatile__("emms")
# elif defined(__MSVC__)
#  define MV_DONE    __asm emms
# else
#  error "Don't know how to insert inline ASM code for this compiler"
# endif
#else
#  define MV_DONE
#endif

/* Copy a frame. */
void MV_COPY_FRAME(void *to, void *from, u32 size);

/* Clear a block. */
void MV_CLEAR_BLOCK(s16 *blk);

   
/* The MV_PRED_?? functions are used to get the predictors from
     the reference block(s). */
void MV_PRED___(s16 *inp, u8 *ref, u32 inc);
void MV_PRED_X_(s16 *inp, u8 *ref, u32 inc);
void MV_PRED__Y(s16 *inp, u8 *ref, u32 inc, u32 inc_y);
void MV_PRED_XY(s16 *inp, u8 *ref, u32 inc, u32 inc_y);


/* The MV_FIELD_PRED_?? functions are similar to the MV_PRED_?? functions,
     except these functions are used when different reference blocks are
     needed for each field. */
void MV_FIELD_PRED___(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 inc);
void MV_FIELD_PRED_X_(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 inc);
void MV_FIELD_PRED__Y(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 inc, u32 inc_y);
void MV_FIELD_PRED_XY(s16 *inp0, u32 inp_inc, 
                      u8  *ref0, u32 inc, u32 inc_y);


/* Saturates the final pixel values, in the block, and stores them in
     the frame buffer. */
void MV_SATURATION(s16 *blk, u8 *outp, int inc);

   
/* Adds the predictor block to the DCT block, to obtain the final pixel
     values.  The final values are stored in the frame buffer. */
void MV_ADD_PRED(s16 *blk, s16 *p_blk, u8 *outp, 
                 int  inc, u32  sh);


/* Copies the predictor block into the frame buffer. */
void MV_COPY_PRED(s16 *p_blk, u8 *outp, int inc, u32 sh);

#endif

