/***************************************************************************
                     MPEG_audio_module.h  -  description
                             -------------------
   MPEG audio PES module functions.
                             -------------------
    begin                : Thu Apr 14 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef __MPEG_AUDIO_MODULE_H__
#define __MPEG_AUDIO_MODULE_H__

#include <stddef.h>
#include "common/typedefs.h"
#include "MPEG2_SYS.h"
#include "PMT_SECTION.h"


/* Use this macro to initialize a MPEG2_PSI_PES_MODULE structure. */
#define MPEG_AUDIO_MODULE \
{\
   MPEG_AUDIO_MODULE_stream_init, \
   MPEG_AUDIO_MODULE_process_pkt, \
   {\
      MPEG_AUDIO_MODULE_mem_needed, \
      MPEG_AUDIO_MODULE_init, \
      MPEG_AUDIO_MODULE_done, \
      MPEG_AUDIO_MODULE_set_offset, \
      MPEG_AUDIO_MODULE_discontinuity, \
      NULL \
   }\
}

#ifdef __cplusplus
extern "C" {
#endif

/*****************************************************************
   GENERAL MODULE FUNCTIONS
     These functions are used by all SYSTEM, PSI, and PES modules.
 *****************************************************************/
/* Returns the number of bytes of memory needed by the module. */
size_t MPEG_AUDIO_MODULE_mem_needed();

/* Initializes the memory provided for the module.
      INPUT: void *     - Pointer to the memory provided for the module.
             SYS_DATA * - Pointer to system data. */
void   MPEG_AUDIO_MODULE_init(void *mem, SYS_DATA *sys_data);

/* Clean up (DO NOT FREE) the memory provided for the module.
      INPUT: void * - Pointer to the memory provided for the module. */
void   MPEG_AUDIO_MODULE_done(void *mem);

/* Gives the module the offset value needed to obtain the program clock value.
      INPUT: void *  - Pointer to the memory provided for the module.
             CLK_VAL - offset value */
void   MPEG_AUDIO_MODULE_set_offset(void *mem, CLK_VAL offset);

/* Alerts the module to a discontinuity in the data stream.
      INPUT: void * - Pointer to the memory provided for the module. */
void   MPEG_AUDIO_MODULE_discontinuity(void *mem);


/**************************************************************************
   PES MODULE FUNCTIONS
     These functions are used by SYSTEM modules to process the  PES_PKTs
       they produce.
 **************************************************************************/
/* Initialize everything needed to handle the given elementary stream.
     The PES module will handle the elementary stream (if able) until the
     MPEG2_MODULE_done function is called, or this function is called
     again.
      INPUT: void *        - Pointer to the memory provided for the module.
             PMT_SECTION * - Pointer to the PMT data.
             int           - Index, in elem array, of the elementary stream
                               that needs to be handled.
      RETURN: Module is ready to handle
                the stream.               - Non-zero value.
              Module cannot handle the
                stream.                   - 0 */
int MPEG_AUDIO_MODULE_stream_init(void *mem, PMT_SECTION *pmt, int idx);

/* Processes the packet handled by the module.
      INPUT: void * - Pointer to the memory provided for the module.
             void * - Pointer to the PES_PKT that needs handling.
                        Set pes->stream_id to zero to let the PES
                        module know to expect the start of a new
                        PES_PKT. */
void MPEG_AUDIO_MODULE_process_pkt(void *mem, void *pes);

#ifdef __cplusplus
};
#endif

#endif

