/***************************************************************************
 VC1_motion_vectors.c
 --------------------
   These are the macros and routines needed to handle the motion vectors.
 --------------------
 begin     : Wed Feb 28 2007
 copyright : (C) 2007 by Dennis Lew
 email     : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define VC1_VIDEO_TRACE*/
#define SEE_INPUT
/*#define SEE_INTERMEDIATE*/
#include <string.h>

#ifdef VC1_VIDEO_TRACE
# include <stdio.h>
#endif

#include "PIC_LAYER.h"


#include "VC1_motion_vectors.h"
static u8 BILINEAR_COEFF[] = {
           /* A   D   B   C */
   /* x y    UL  UR  LL  LR */
   /* 0 1 */ 12,  0,  4,  0,
   /* 0 2 */  8,  0,  8,  0,
   /* 0 3 */  4,  0, 12,  0,
   /* 1 0 */ 12,  4,  0,  0,
   /* 1 1 */  9,  3,  3,  1,
   /* 1 2 */  6,  2,  6,  2,
   /* 1 3 */  3,  1,  9,  3,
   /* 2 0 */  8,  8,  0,  0,
   /* 2 1 */  6,  6,  2,  2,
   /* 2 2 */  4,  4,  4,  4,
   /* 2 3 */  2,  2,  6,  6,
   /* 3 0 */  4, 12,  0,  0,
   /* 3 1 */  3,  9,  1,  3,
   /* 3 2 */  2,  6,  2,  6,
   /* 3 3 */  1,  3,  3,  9
},

          BICUBIC_COEFF1[] = {
   /* x y    y     x */
   /* 0 1 */ 6, 0, 0, 0,
   /* 0 2 */ 4, 4, 0, 0,
   /* 0 3 */ 6, 8, 0, 0,
   /* 1 0 */ 0, 0, 6, 0,
   /* 1 1 */ 5, 0, 7, 0,
   /* 1 2 */ 3, 4, 7, 0,
   /* 1 3 */ 5, 8, 7, 0,
   /* 2 0 */ 0, 0, 4, 4,
   /* 2 1 */ 3, 0, 7, 4,
   /* 2 2 */ 1, 4, 7, 4,
   /* 2 3 */ 3, 8, 7, 4,
   /* 3 0 */ 0, 0, 6, 8,
   /* 3 1 */ 5, 0, 7, 8,
   /* 3 2 */ 3, 4, 7, 8,
   /* 3 3 */ 5, 8, 7, 8
};
static s16 BICUBIC_COEFF2[] = {
   /*        P1  P2  P3  P4 */
   /* 1/4 */ -4, 53, 18, -3,
   /* 1/2 */ -1,  9,  9, -1,
   /* 3/4 */ -3, 18, 53, -4
};

s16 VC1_median4(s16 a, s16 b, s16 c, s16 d)
{
   s16 max = a,
       min = a;

   a += b;
   if (max < b)
      max = b;
   else if (min > b)
      min = b;
   a += c;
   if (max < c)
      max = c;
   else if (min > c)
      min = c;
   a += d;
   if (max < d)
      max = d;
   else if (min > d)
      min = d;
   return ((a - max - min) / 2);
}

s16 VC1_median3(s16 a, s16 b, s16 c)
{
   if (a > b) {
      if (b > c)
         return b;
      else if (a > c)
         return c;
      else
         return a;
   }
   else if (a > c)
      return a;
   else if (b > c)
      return c;
   else
      return b;
}

/* pg 174-175: Calculate UV vector */
/* pg 183-188: Motion vector operations */

void VC1_1MVinc(s32 *MVinc, u16 *MVout_qp, s16 *MV,
                u16 x, u16 y, u16 wid, u16 hgt, u32 frame_wid, u8 fastuv)
{
   s16 mvx, mvy;

   MVinc[1] =
   MVinc[2] =
   MVinc[3] = 0;

   /* calculate luma MVinc value */
   MVout_qp[0] = (u16)(MV[0] & 3) << 2;
   mvx         = (s16)x + (MV[0] >> 2);
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("      Y_pos  = %hd,", mvx);
   VC1_TRACE_END_BLK
#endif
   if (mvx <= 0) {
      if (mvx <= -8) {
         MVout_qp[2] = (mvx == -8) ? 0x4800 : 0x5000;
         if (mvx <= -17)
            MVout_qp[3] = 0x5000;
         else
            MVout_qp[3] = (u16)(-7 - mvx) << 11;
         MVinc[1] -= 8;
         MVinc[3] -= 8;
      }
      else {
         MVout_qp[3] = 0;
         MVout_qp[2] = (u16)(1 - mvx) << 11;
         MVinc[1]   -= (s32)(1 - mvx);
         MVinc[3]   -= (s32)(1 - mvx);
      }
      mvx = 0;
   }
   else if (mvx >= (s16)(wid-17)) {
      if (mvx >= (s16)(wid-9)) {
         MVout_qp[3] = (mvx == (s16)(wid-9)) ? 0xC800 : 0xD000;
         if (mvx >= (s16)wid) {
            mvx         = (s16)wid;
            MVout_qp[2] = 0xD000;
            MVinc[1]   -= 8;
            MVinc[3]   -= 8;
         }
         else {
            MVout_qp[2]  =     (((u16)mvx - wid + 10) << 11) | 0x8000;
            if (mvx > (s16)(wid-8)) {
               MVinc[1] -= (s32)((u16)mvx - wid +  8);
               MVinc[3] -= (s32)((u16)mvx - wid +  8);
            }
         }
      }
      else {
         MVout_qp[2] = 0;
         MVout_qp[3] = (((u16)mvx - wid + 18) << 11) | 0x8000;
      }
      mvx--;
   }
   else {
      MVout_qp[2] = 0;
      MVout_qp[3] = 0;
      mvx--;
   }
   mvx -= (s16)x;
   MVout_qp[0] |= MVout_qp[2];
   MVout_qp[1]  = MVout_qp[3];

   MVout_qp[0] |= (u16)(MV[1] & 3);
   mvy          = (s16)y + (MV[1] >> 2);
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("%hd\n", mvy);
   VC1_TRACE_END_BLK
#endif
   if (mvy <= 0) {
      if (mvy <= -8) {
         MVout_qp[1] |= (mvy == -8) ? 0x0240 : 0x0280;
         if (mvy <= -17)
            MVout_qp[3] |= 0x0280;
         else
            MVout_qp[3] |= (u16)(-7 - mvy) << 6;
         MVinc[2] -= (s32)frame_wid << 3;
         MVinc[3] -= (s32)frame_wid << 3;
      }
      else {
         MVout_qp[1] |= (u16)(1 - mvy) << 6;
         MVinc[2]    -= (s32)(1 - mvy) * (s32)frame_wid;
         MVinc[3]    -= (s32)(1 - mvy) * (s32)frame_wid;
      }
      mvy = 0;
   }
   else if (mvy >= (s16)(hgt-17)) {
      if (mvy >= (s16)(hgt-9)) {
         MVout_qp[3] |= (mvy == (s16)(hgt-9)) ? 0x0640 : 0x0680;
         if (mvy >= (s16)hgt) {
            mvy          = (s16)hgt;
            MVout_qp[1] |= 0x0680;
            MVinc[2]    -= (s32)frame_wid << 3;
            MVinc[3]    -= (s32)frame_wid << 3;
         }
         else {
            MVout_qp[1] |=     (((u16)mvy - hgt + 10) << 6) | 0x0400;
            if (mvy > (s16)(hgt-8)) {
               MVinc[2] -= (s32)((u16)mvy - hgt +  8) * (s32)frame_wid;
               MVinc[3] -= (s32)((u16)mvy - hgt +  8) * (s32)frame_wid;
            }
         }
      }
      else
         MVout_qp[3] |= (((u16)mvy - hgt + 18) << 6) | 0x0400;
      mvy--;
   }
   else
      mvy--;
   mvy -= (s16)y;
   MVout_qp[0] |= 0x07C0 & MVout_qp[1];
   MVout_qp[2] |= 0x07C0 & MVout_qp[3];
   MVout_qp[1] |= 0x000F & MVout_qp[0];
   MVout_qp[2] |= 0x000F & MVout_qp[0];
   MVout_qp[3] |= 0x000F & MVout_qp[0];

   MVinc[0]  = (s32)mvx + ((s32)mvy * (s32)frame_wid);
   MVinc[1] += MVinc[0];
   MVinc[2] += MVinc[0];
   MVinc[3] += MVinc[0];

   /* calculate chroma MV */
   mvx = MV[0];
   if ((MV[0] & 3) == 3)
      mvx++;
   mvx >>= 1;
   mvy = MV[1];
   if ((MV[1] & 3) == 3)
      mvy++;
   mvy >>= 1;
   if (fastuv) {
      if (mvx & 1) {
         if (mvx > 0)
            mvx--;
          else
            mvx++;
      }
      if (mvy & 1) {
         if (mvy > 0)
            mvy--;
         else
            mvy++;
      }
   }

   /* calculate chroma MVinc value */
   x         >>= 1;
   y         >>= 1;
   wid       >>= 1;
   hgt       >>= 1;
   frame_wid >>= 1;
   MVout_qp[4] = (u16)(mvx & 3) << 2;
   mvx       >>= 2;
   mvx        += (s16)x;
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("      C_pos  = %hd,", mvx);
   VC1_TRACE_END_BLK
#endif
   if (mvx <= 0) {
      MVout_qp[4] |= (mvx <= -9) ? 0x5000 : ((u16)(1 - mvx) << 11);
      mvx          = 0;
   }
   else {
      if (mvx >= (s16)wid) {
         mvx          = (s16)wid;
         MVout_qp[4] |= 0xD000;
      }
      else if (mvx >= (s16)(wid-9))
         MVout_qp[4] |= (((u16)mvx - wid + 10) << 11) | 0x8000;
      mvx--;
   }
   mvx -= (s16)x;

   MVout_qp[4] |= (u16)(mvy & 3);
   mvy        >>= 2;
   mvy         += (s16)y;
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("%hd\n", mvy);
   VC1_TRACE_END_BLK
#endif
   if (mvy <= 0) {
      MVout_qp[4] |= (mvy <= -9) ? 0x0280 : ((u16)(1 - mvy) << 6);
      mvy          = 0;
   }
   else {
      if (mvy >= (s16)hgt) {
         mvy          = (s16)hgt;
         MVout_qp[4] |= 0x0680;
      }
      else if (mvy >= (s16)(hgt-9))
         MVout_qp[4] |= (((u16)mvy - hgt + 10) << 6) | 0x0400;
      mvy--;
   }
   mvy -= (s16)y;

   MVinc[4]    = (s32)mvx + ((s32)mvy * (s32)frame_wid);
   MVinc[5]    = MVinc[4];
   MVout_qp[5] = MVout_qp[4];

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("      inc    = %d %d %d %d %d\n"
          "      out_qp = 0x%04hX 0x%04hX 0x%04hX 0x%04hX 0x%04hX\n",
          MVinc[0], MVinc[1], MVinc[2], MVinc[3], MVinc[4],
          MVout_qp[0], MVout_qp[1], MVout_qp[2], MVout_qp[3], MVout_qp[4]);
   VC1_TRACE_END_BLK
#endif
}

void VC1_4MVinc(s32 *MVinc, u16 *MVout_qp, s16 *MV,
                u16 x, u16 y, u16 wid, u16 hgt,
                u32 frame_wid, u8 fastuv, u8 *MV4mode)
{
   s16 mvx, mvy;
   s16 i      = 0,
       j      = 0,
       idx[4] = {-1, -1, -1, -1};
   u8  intra  = *MV4mode;

   do {
      if (intra < 0x80) {
         /* calculate luma MVinc value */
         idx[j++]  = i;
         *MVout_qp = (u16)(MV[i] & 3) << 2;
         mvx       = (s16)x + (MV[i] >> 2) + ((i & 2) << 2);
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("        Y_pos[%hd]  = %hd,", (i>>1), mvx);
         VC1_TRACE_END_BLK
#endif
         if (mvx <= 0) {
            *MVout_qp |= (mvx <= -9) ? 0x5000 : ((u16)(1 - mvx) << 11);
            mvx        = 0;
         }
         else {
            if (mvx >= (s16)wid) {
               mvx        = (s16)wid;
               *MVout_qp |= 0xD000;
            }
            else if (mvx >= (s16)(wid-9))
               *MVout_qp |= (((u16)mvx - wid + 10) << 11) | 0x8000;
            mvx--;
         }
         mvx -= (s16)x + ((i & 2) << 2);

         i++;
         *MVout_qp |= (u16)(MV[i] & 3);
         mvy        = (s16)y + (MV[i] >> 2) + ((i & 4) << 1);
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf("%hd\n", mvy);
         VC1_TRACE_END_BLK
#endif
         if (mvy <= 0) {
            *MVout_qp |= (mvy <= -9) ? 0x0280 : ((u16)(1 - mvy) << 6);
            mvy        = 0;
         }
         else {
            if (mvy >= (s16)hgt) {
               mvy        = (s16)hgt;
               *MVout_qp |= 0x0680;
            }
            else if (mvy >= (s16)(hgt-9))
               *MVout_qp |= (((u16)mvy - hgt + 10) << 6) | 0x0400;
            mvy--;
         }
         mvy -= (s16)y + ((i & 4) << 1);

         *MVinc = (s32)mvx + ((s32)mvy * (s32)frame_wid);
      }
      else
         i++;
      MVinc++;
      MVout_qp++;
      intra <<= 1;
   } while (++i < 8);

   if (j < 2)
      *MV4mode |= 0x0C;    /* chroma blocks are intra */
   else {
      /* calculate chroma MV */
      if (j == 4) {
         mvx = VC1_median4(MV[idx[0]],   MV[idx[1]],   MV[idx[2]],   MV[idx[3]]);
         mvy = VC1_median4(MV[idx[0]+1], MV[idx[1]+1], MV[idx[2]+1], MV[idx[3]+1]);
      }
      else if (j == 3) {
         mvx = VC1_median3(MV[idx[0]],   MV[idx[1]],   MV[idx[2]]);
         mvy = VC1_median3(MV[idx[0]+1], MV[idx[1]+1], MV[idx[2]+1]);
      }
      else {
         mvx = (MV[idx[0]] +   MV[idx[1]])   / 2;
         mvy = (MV[idx[0]+1] + MV[idx[1]+1]) / 2;
      }

      if ((mvx & 3) == 3)
         mvx++;
      mvx >>= 1;
      if ((mvy & 3) == 3)
         mvy++;
      mvy >>= 1;
      if (fastuv) {
         if (mvx & 1) {
            if (mvx > 0)
               mvx--;
            else
               mvx++;
         }
         if (mvy & 1) {
            if (mvy > 0)
               mvy--;
            else
               mvy++;
         }
      }

      /* calculate chroma MVinc value */
      x         >>= 1;
      y         >>= 1;
      wid       >>= 1;
      hgt       >>= 1;
      frame_wid >>= 1;
      *MVout_qp   = (u16)(mvx & 3) << 2;
      mvx       >>= 2;
      mvx        += (s16)x;
#ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("        C_pos     = %hd,", mvx);
      VC1_TRACE_END_BLK
#endif
      if (mvx <= 0) {
         *MVout_qp |= (mvx <= -9) ? 0x5000 : ((u16)(1 - mvx) << 11);
         mvx        = 0;
      }
      else {
         if (mvx >= (s16)wid) {
            mvx        = (s16)wid;
            *MVout_qp |= 0xD000;
         }
         else if (mvx >= (s16)(wid-9))
            *MVout_qp |= (((u16)mvx - wid + 10) << 11) | 0x8000;
         mvx--;
      }
      mvx -= (s16)x;

      *MVout_qp |= (u16)(mvy & 3);
      mvy      >>= 2;
      mvy       += (s16)y;
#ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("%hd\n", mvy);
      VC1_TRACE_END_BLK
#endif
      if (mvy <= 0) {
         *MVout_qp |= (mvy <= -9) ? 0x0280 : ((u16)(1 - mvy) << 6);
         mvy        = 0;
      }
      else {
         if (mvy >= (s16)hgt) {
            mvy        = (s16)hgt;
            *MVout_qp |= 0x0680;
         }
         else if (mvy >= (s16)(hgt-9))
            *MVout_qp |= (((u16)mvy - hgt + 10) << 6) | 0x0400;
         mvy--;
      }
      mvy -= (s16)y;

      *MVinc      = (s32)mvx + ((s32)mvy * (s32)frame_wid);
      MVinc[1]    = *MVinc;
      MVout_qp[1] = *MVout_qp;
   }
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   MVinc    -= 4;
   MVout_qp -= 4;
   printf("        inc    = %d %d %d %d",
          MVinc[0], MVinc[1], MVinc[2], MVinc[3]);
   if (j >= 2)
      printf(" %d", MVinc[4]);
   printf("\n        out_qp = 0x%04hX 0x%04hX 0x%04hX 0x%04hX",
          MVout_qp[0], MVout_qp[1], MVout_qp[2], MVout_qp[3]);
   if (j >= 2)
      printf(" 0x%04hX\n", MVout_qp[4]);
   else
      printf(" intra\n");
   VC1_TRACE_END_BLK
#endif
}


VC1_INT_FUNC VC1_get_blk(u8 *blk, u8 *ref, u16 MVout, u32 frame_wid,
                         u8 MVmode, u8 *LUT)
{
   VC1_INT_FUNC func;
   u16 col,
       row    =  0,
       dup_x  =  0,
       cpy_x  = 11,
       size_x = 11,
       dup_y  =  0,
       end_y  = 11,
       size_y = 11;

   if (MVout & 0x8000)
      cpy_x -= (MVout >> 11) & 0x000F; /* trailing horiz duplicate pixels */
   else if (MVout & 0x7800) {
      dup_x  =  MVout >> 11;           /* leading horiz duplicate pixels */
      cpy_x -= dup_x;
   }

   if (MVout & 0x0400)
      end_y -= (MVout >> 6) & 0x000F;  /* trailing vert duplicate rows */
   else if (MVout & 0x03C0)
      dup_y  = (MVout >> 6) & 0x000F;  /* leading vert duplicate rows */

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("      dup_x=%2hu  cpy_x=%2hu  size_x=%2hu"
          "      dup_y=%2hu  end_y=%2hu  size_y=%2hu\n",
          dup_x, cpy_x, size_x,
          dup_y, end_y, size_y);
   VC1_TRACE_END_BLK
#endif

   MVout &= 0x000F;  /* qp location */
   if (!MVout) {
      /* only need an 8x8 block */
      func = NULL;
      if (dup_x) {
         if (dup_x > 8) {
            dup_x = 7;
            cpy_x = 1;
         }
         else {
            dup_x--;
            cpy_x -= 2;
         }
      }
      else if (cpy_x > 1) {
         if (cpy_x > 8)
            cpy_x = 8;
         else
            cpy_x--;
         ref++;
      }
      size_x = 8;

      if (dup_y) {
         if (dup_y > 8)
            dup_y = 7;
         else
            dup_y--;
      }
      else if (end_y > 1) {
         if (end_y > 8)
            end_y = 8;
         else
            end_y--;
         ref += frame_wid;
      }
      size_y = 8;
   }
   else if (MVmode & VC1_MVMODE_BILINEAR) {
      /* only need an 9x9 block */
      func = VC1_MV_BILINEAR;
      if (dup_x) {
         if (dup_x > 9) {
            dup_x = 8;
            cpy_x = 1;
         }
         else {
            dup_x--;
            cpy_x--;
         }
      }
      else if (cpy_x > 1) {
         if (cpy_x > 9)
            cpy_x = 9;
         else
            cpy_x--;
         ref++;
      }
      size_x = 9;

      if (dup_y) {
         if (dup_y > 9)
            dup_y = 8;
         else
            dup_y--;
      }
      else if (end_y > 1) {
         if (end_y > 9)
            end_y = 9;
         else
            end_y--;
         ref += frame_wid;
      }
      size_y = 9;
   }
   else if (MVout <= 0x0003) {
      /* only need an 8x11 block */
      func = VC1_MV_BICUBIC__Y;
      if (dup_x) {
         if (dup_x > 8) {
            dup_x = 7;
            cpy_x = 1;
         }
         else {
            dup_x--;
            cpy_x -= 2;
         }
      }
      else if (cpy_x > 1) {
         if (cpy_x > 8)
            cpy_x = 8;
         else
            cpy_x--;
         ref++;
      }
      size_x = 8;
   }
   else if (!(MVout & 0x0003)) {
      /* only need a 11x8 block */
      func = VC1_MV_BICUBIC_X_;
      if (dup_y) {
         if (dup_y > 8)
            dup_y = 7;
         else
            dup_y--;
      }
      else if (end_y > 1) {
         if (end_y > 8)
            end_y = 8;
         else
            end_y--;
         ref += frame_wid;
      }
      size_y = 8;
   }
   else
      func = VC1_MV_BICUBIC_XY;

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("      dup_x=%2hu  cpy_x=%2hu  size_x=%2hu"
          "      dup_y=%2hu  end_y=%2hu  size_y=%2hu\n",
          dup_x, cpy_x, size_x,
          dup_y, end_y, size_y);
   if (!func)
      printf("  INPUT\n");
   VC1_TRACE_END_BLK
#endif

   if (MVmode & VC1_MVMODE_INT_COMP) {
      do {
         col = 0;

         /* leading horiz duplicate pixels */
         while (col < dup_x) {
            blk[col++] = LUT[ref[0]];
         }

         /* copy reference pixels */
         while (col < (dup_x+cpy_x)) {
            blk[col] = LUT[ref[col-dup_x]];
            col++;
         }

         /* trailing horiz duplicate pixels */
         while (col < size_x) {
            blk[col++] = LUT[ref[cpy_x-1]];
         }
#if defined(VC1_VIDEO_TRACE) && defined(SEE_INPUT)
         VC1_TRACE_BEG_BLK
            if (!func) {
               col = 0;
               do {
                  printf(" %02hX", (u16)blk[col]);
               } while (++col < size_x);
               printf("\n");
            }
         VC1_TRACE_END_BLK
#endif
         blk += size_x;
         ref += frame_wid;

         /* leading vert duplicate rows */
         while (row < dup_y) {
#if defined(VC1_VIDEO_TRACE) && defined(SEE_INPUT)
            VC1_TRACE_BEG_BLK
               if (!func) {
                  col = 0;
                  do {
                     printf(" %02hX", (u16)blk[col]);
                  } while (++col < size_x);
                  printf("\n");
               }
            VC1_TRACE_END_BLK
#endif
            memcpy(blk, (blk-size_x), (size_t)size_x);
            blk += size_x;
            row++;
         }
      } while (++row < end_y);
   }
   else {
      do {
         col = 0;

         /* leading horiz duplicate pixels */
         while (col < dup_x) {
            blk[col++] = ref[0];
         }

         /* copy reference pixels */
         memcpy((blk+col), ref, (size_t)cpy_x);
         col += cpy_x;

         /* trailing horiz duplicate pixels */
         while (col < size_x) {
            blk[col++] = ref[cpy_x-1];
         }
#if defined(VC1_VIDEO_TRACE) && defined(SEE_INPUT)
         VC1_TRACE_BEG_BLK
            if (!func) {
               col = 0;
               do {
                  printf(" %02hX", (u16)blk[col]);
               } while (++col < size_x);
               printf("\n");
            }
         VC1_TRACE_END_BLK
#endif
         blk += size_x;
         ref += frame_wid;

         /* leading vert duplicate rows */
         while (row < dup_y) {
#if defined(VC1_VIDEO_TRACE) && defined(SEE_INPUT)
            VC1_TRACE_BEG_BLK
               if (!func) {
                  col = 0;
                  do {
                     printf(" %02hX", (u16)blk[col]);
                  } while (++col < size_x);
                  printf("\n");
               }
            VC1_TRACE_END_BLK
#endif
            memcpy(blk, (blk-size_x), (size_t)size_x);
            blk += size_x;
            row++;
         }
      } while (++row < end_y);
   }
   /* trailing vert duplicate rows */
   while (row < size_y) {
#if defined(VC1_VIDEO_TRACE) && defined(SEE_INPUT)
      VC1_TRACE_BEG_BLK
         if (!func) {
            col = 0;
            do {
               printf(" %02hX", (u16)blk[col]);
            } while (++col < size_x);
            printf("\n");
         }
      VC1_TRACE_END_BLK
#endif
      memcpy(blk, (blk-size_x), (size_t)size_x);
      blk += size_x;
      row++;
   }
   return func;
}

void VC1_MV_BILINEAR(u8 *blk, u16 MV_qp, u8 rnd)
{
   const
   u8  *coeff;
   u8  *out;
   u16  tmp, col, row;

   coeff = (BILINEAR_COEFF-4) + (MV_qp << 2);
   rnd   = 8 - rnd;
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("BILINEAR()\n"
          "  UL=%hd LL=%hd UR=%hd LR=%hd R=%hd shift=4\n",
          coeff[0], coeff[2], coeff[1], coeff[3], (s16)rnd);
   VC1_TRACE_END_BLK
#endif
   out = blk;
#ifdef VC1_VIDEO_TRACE
# ifdef SEE_INPUT
   VC1_TRACE_BEG_BLK
   printf("  INPUT\n");
   col = 9;
   do {
      printf(" %02hX", (u16)blk[0]);
      blk++;
   } while (--col);
   printf("\n");
   blk = out;
   VC1_TRACE_END_BLK
# endif
#endif
   row = 8;
   do {
      col = 8;
      do {
#ifdef VC1_VIDEO_TRACE
# ifdef SEE_INPUT
         VC1_TRACE_BEG_BLK
         printf(" %02hX", (u16)blk[9]);
         VC1_TRACE_END_BLK
# endif
#endif
         tmp = (u16)rnd +
               ((u16)blk[ 0] * (u16)coeff[0]) +
               ((u16)blk[ 1] * (u16)coeff[1]) +
               ((u16)blk[ 9] * (u16)coeff[2]) +
               ((u16)blk[10] * (u16)coeff[3]);
         *(out++) = (u8)(tmp >> 4);
         blk++;
      } while (--col);
#ifdef VC1_VIDEO_TRACE
# ifdef SEE_INPUT
      VC1_TRACE_BEG_BLK
      printf(" %02hX\n", (u16)blk[9]);
      VC1_TRACE_END_BLK
# endif
#endif
      blk++;
   } while (--row);
}

void VC1_MV_BICUBIC_X_(u8 *blk, u16 MV_qp, u8 rnd)
{
   const
   u8  *coeff1;
   const
   s16 *coeff2;
   u8  *out;
   s16  tmp, max, rv, col, row;

   coeff1 = (BICUBIC_COEFF1-4) + (MV_qp << 2);
   /* horizontal interpolation */
   rv     = (1 << (coeff1[2]-1)) - (s16)rnd;
   coeff2 = BICUBIC_COEFF2 + coeff1[3];
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("BICUBIC_X_()\n"
          "  F0=%hd F1=%hd F2=%hd F3=%hd R=%hd shift=%hd\n",
          coeff2[0], coeff2[1], coeff2[2], coeff2[3], rv, (s16)coeff1[2]);
   VC1_TRACE_END_BLK
#endif
   out = blk;
   max = 255 << coeff1[2];
   row = 8;
   do {
#ifdef VC1_VIDEO_TRACE
# ifdef SEE_INPUT
      VC1_TRACE_BEG_BLK
      if (row == 8)
         printf("  INPUT\n");
      col = 3;
      do {
         printf(" %02hX", (u16)blk[0]);
         blk++;
      } while (--col);
      blk -= 3;
      VC1_TRACE_END_BLK
# endif
#endif
      col = 8;
      do {
#ifdef VC1_VIDEO_TRACE
# ifdef SEE_INPUT
         VC1_TRACE_BEG_BLK
         printf(" %02hX", (u16)blk[3]);
         VC1_TRACE_END_BLK
# endif
#endif
         tmp = rv +
               ((s16)blk[0] * coeff2[0]) +
               ((s16)blk[1] * coeff2[1]) +
               ((s16)blk[2] * coeff2[2]) +
               ((s16)blk[3] * coeff2[3]);
         if (tmp <= 0)
            *out = 0;
         else if (tmp >= max)
            *out = 255;
         else
            *out = (u8)(tmp >> coeff1[2]);
         out++;
         blk++;
      } while (--col);
      blk += 3;
#ifdef VC1_VIDEO_TRACE
# ifdef SEE_INPUT
      VC1_TRACE_BEG_BLK
      printf("\n");
      VC1_TRACE_END_BLK
# endif
#endif
   } while (--row);
}

void VC1_MV_BICUBIC__Y(u8 *blk, u16 MV_qp, u8 rnd)
{
   const
   u8  *coeff1;
   const
   s16 *coeff2;
   u8  *out;
   s16  tmp, max, rv, col, row;

   coeff1 = (BICUBIC_COEFF1-4) + (MV_qp << 2);
   /* vertical interpolation */
   rv     = (1 << (coeff1[0]-1)) - 1 + (s16)rnd;
   coeff2 = BICUBIC_COEFF2 + coeff1[1];
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("BICUBIC__Y()\n"
          "  F0=%hd F1=%hd F2=%hd F3=%hd R=%hd shift=%hd\n",
          coeff2[0], coeff2[1], coeff2[2], coeff2[3], rv, (s16)coeff1[0]);
   VC1_TRACE_END_BLK
#endif
   out = blk;
#ifdef VC1_VIDEO_TRACE
# ifdef SEE_INPUT
   VC1_TRACE_BEG_BLK
   printf("  INPUT\n");
   row = 3;
   do {
      col = 8;
      do {
         printf(" %02hX", (u16)blk[0]);
         blk++;
      } while (--col);
      printf("\n");
   } while (--row);
   blk = out;
   VC1_TRACE_END_BLK
# endif
#endif
   max = 255 << coeff1[0];
   row = 8;
   do {
      col = 8;
      do {
#ifdef VC1_VIDEO_TRACE
# ifdef SEE_INPUT
         VC1_TRACE_BEG_BLK
         printf(" %02hX", (u16)blk[24]);
         VC1_TRACE_END_BLK
# endif
#endif
         tmp = rv +
               ((s16)blk[ 0] * coeff2[0]) +
               ((s16)blk[ 8] * coeff2[1]) +
               ((s16)blk[16] * coeff2[2]) +
               ((s16)blk[24] * coeff2[3]);
         if (tmp <= 0)
            *out = 0;
         else if (tmp >= max)
            *out = 255;
         else
            *out = (u8)(tmp >> coeff1[0]);
         out++;
         blk++;
      } while (--col);
#ifdef VC1_VIDEO_TRACE
# ifdef SEE_INPUT
      VC1_TRACE_BEG_BLK
      printf("\n");
      VC1_TRACE_END_BLK
# endif
#endif
   } while (--row);
}

void VC1_MV_BICUBIC_XY(u8 *blk, u16 MV_qp, u8 rnd)
{
   const
   u8  *coeff1;
   const
   s16 *coeff2x, *coeff2y;
   u8  *out;
   s16  rvx, rvy, col, row;
   s32  tmp2, max;
   s16  tmp[11];

   coeff1  = (BICUBIC_COEFF1-4) + (MV_qp << 2);
   rvy     = (1 << (coeff1[0]-1)) - 1 + (s16)rnd;
   coeff2y = BICUBIC_COEFF2 + coeff1[1];
   rvx     = (1 << (coeff1[2]-1)) - (s16)rnd;
   coeff2x = BICUBIC_COEFF2 + coeff1[3];

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("BICUBIC_XY()\n"
          "  VERT> F0=%hd F1=%hd F2=%hd F3=%hd R=%hd shift=%hd\n",
          coeff2y[0], coeff2y[1], coeff2y[2], coeff2y[3], rvy, (s16)coeff1[0]);
   VC1_TRACE_END_BLK
#endif
   out = blk;
#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
# ifdef SEE_INPUT
   printf("  INPUT\n");
   row = 3;
   do {
      col = 11;
      do {
         printf(" %02hX", (u16)blk[0]);
         blk++;
      } while (--col);
      printf("\n");
   } while (--row);
   blk = out;
# else
#  ifdef SEE_INTERMEDIATE
   printf("  INTERMEDIATE\n");
#  endif
# endif
   VC1_TRACE_END_BLK
#endif
   max = 255 << coeff1[2];
   row = 8;
   do {
      /* vertical interpolation */
      col = 0;
      do {
         tmp[col] = rvy +
                    ((s16)blk[ 0] * coeff2y[0]) +
                    ((s16)blk[11] * coeff2y[1]) +
                    ((s16)blk[22] * coeff2y[2]) +
                    ((s16)blk[33] * coeff2y[3]);
         tmp[col] >>= coeff1[0];
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
# ifdef SEE_INPUT
         printf(" %02hX", (u16)blk[33]);
# else
#  ifdef SEE_INTERMEDIATE
         printf("%6hd", tmp[col]);
#  endif
# endif
         VC1_TRACE_END_BLK
#endif
         blk++;
      } while (++col < 11);
#if defined(VC1_VIDEO_TRACE) && (defined(SEE_INPUT)||defined(SEE_INTERMEDIATE))
      VC1_TRACE_BEG_BLK
      printf("\n");
      VC1_TRACE_END_BLK
#endif

      /* horizontal interpolation */
      col = 0;
      do {
         tmp2 = (s32)rvx +
                (s32)(tmp[col]   * coeff2x[0]) +
                (s32)(tmp[col+1] * coeff2x[1]) +
                (s32)(tmp[col+2] * coeff2x[2]) +
                (s32)(tmp[col+3] * coeff2x[3]);
         if (tmp2 <= 0)
            *out = 0;
         else if (tmp2 >= max)
            *out = 255;
         else
            *out = (u8)(tmp2 >> coeff1[2]);
         out++;
      } while (++col < 8);
   } while (--row);

#ifdef VC1_VIDEO_TRACE
   VC1_TRACE_BEG_BLK
   printf("  HORIZ> F0=%hd F1=%hd F2=%hd F3=%hd R=%hd shift=%hd\n",
          coeff2x[0], coeff2x[1], coeff2x[2], coeff2x[3], rvx, (s16)coeff1[2]);
   VC1_TRACE_END_BLK
#endif
}

void VC1_MV_ADD_PRED(s16 *err, u8 *blk, u8 *outp, int inc)
{
   u16 col,
       row = 8;

   inc -= 8;
   do {
      col = 8;
      do {
         *err += (s16)(*blk);
         if (*err < 0)
            *outp = 0;
         else if (*err > 255)
            *outp = 255;
         else
            *outp = (u8)(*err);
         err++;
         blk++;
         outp++;
      } while (--col);
      outp += inc;
   } while (--row);
}

#ifdef VC1_VIDEO_TRACE
void VC1_MV_COPY_OUTPUT(s16 *err, u8 *outp, int inc)
{
   u16 col,
       row = 8;

   inc -= 8;
   do {
      col = 8;
      do {
         *err = (s16)(*outp);
         err++;
         outp++;
      } while (--col);
      outp += inc;
   } while (--row);
}
#endif

void VC1_MV_BUILD_LUT_TABLE(u8 *tab, s32 scale, s32 shift)
{
   s32 i;

   if (!scale) {
      scale  =  -1*64;
      shift *=  -2*64;
      shift += 255*64;
      if (shift < ((255 - (2*31))*64))
         shift += 128*64;
   }
   else {
      scale += 32;
      if (shift > 31)
         shift -= 64;
      shift <<= 6;
   }

   shift += 32;
   i      = 256;
   do {
      if (shift <= 0)
         *tab = 0;
      else if (shift >= (255 << 6))
         *tab = 255;
      else
         *tab = (u8)(shift >> 6);
      shift += scale;
      tab++;
   } while (--i);

   shift = ((128*64) + 32) - (scale << 7);
   i     = 256;
   do {
      if (shift <= 0)
         *tab = 0;
      else if (shift >= (255 << 6))
         *tab = 255;
      else
         *tab = (u8)(shift >> 6);
      shift += scale;
      tab++;
   } while (--i);
}

