/***************************************************************************
                         bitplane.c  -  description
                             -------------------
   Stuff needed for handling BITPLANEs.
                             -------------------
    begin                : Mon Jun 12 2006
    copyright            : (C) 2006 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define _BITSTREAM_TRACE_
#define BIT_TRACE_BEG_BLK  VC1_TRACE_BEG_BLK
#define BIT_TRACE_END_BLK  VC1_TRACE_END_BLK*/

/*#define VC1_VIDEO_TRACE 2*/
#ifdef VC1_VIDEO_TRACE
# include <stdio.h>
# include "PIC_LAYER.h"
#endif
#include <string.h>

#include "bitplane.h"
#include "bitstream.h"

#define DATA_PTR        ptr
#define END_DATA_PTR    end_pos
#define PARSER_STATE    seq->BPstate


/* Read the BITPLANE data.
      INPUT: seq - Pointer to a SEQ_LAYER structure.
             bp  - Pointer to the BITPLANE.
      RETURNS: non-zero value when entire BITPLANE is complete. */
u32 VC1_VIDEO_read_BITPLANE(SEQ_LAYER *seq, u8 *bp)
{
   const u8 *ptr       = seq->payload,
            *end_pos   = seq->payload + seq->payload_len;
   u8       *bp_start  = bp;
   u16       MBcols    = seq->wid >> 4,
             MBrows    = seq->hgt >> 4;
   u8        bit_shift = seq->bit_shift,
             data      = seq->data,
             bit_shift_inc,
             tmp1;

#undef  DATA_BYTE
#define DATA_BYTE       data
#undef  BIT_SHIFT_BYTE
#define BIT_SHIFT_BYTE  bit_shift
#undef  STOP_PARSING
#define STOP_PARSING    goto VC1_VIDEO_read_BITPLANE_END

#define BITPLANE_MODE_COLSKIP    0x0002
#define BITPLANE_MODE_ROWSKIP    0x0004
#define BITPLANE_MODE_NORM2      0x0008
#define BITPLANE_MODE_NORM6      0x0010
#define BITPLANE_MODE_NORM6_CS1  0x0020
#define BITPLANE_MODE_NORM6_CS2  0x0040
#define BITPLANE_MODE_NORM6_RS1  0x0080
#define BITPLANE_MODE_NORM6_32   0x0100
#define BITPLANE_DIFF            0x0200
#define BITPLANE_INVERT          0x0400

   if (!PARSER_STATE) {
      seq->BPcnt = 0;

      GET_LSB;
      bit_shift_inc = 1;
      if (DATA_BYTE >= 0x80)
         PARSER_STATE = BITPLANE_INVERT;

      DATA_BYTE &= 0x7F;
      if (DATA_BYTE >= 0x60) {
         bit_shift_inc += 2;
         PARSER_STATE  |= BITPLANE_MODE_NORM6;
         seq->BPcnt2    = MBcols;
      }
      else if (DATA_BYTE >= 0x40) {
         bit_shift_inc += 2;
         PARSER_STATE  |= BITPLANE_MODE_NORM2;
      }
      else if (DATA_BYTE >= 0x30) {
         bit_shift_inc += 3;
         PARSER_STATE  |= BITPLANE_MODE_COLSKIP;
         seq->BPcnt     = MBcols;
      }
      else if (DATA_BYTE >= 0x20) {
         bit_shift_inc += 3;
         PARSER_STATE  |= BITPLANE_MODE_ROWSKIP;
         seq->BPcnt     = MBrows;
      }
      else if (DATA_BYTE >= 0x10) {
         bit_shift_inc += 3;
         PARSER_STATE  |= BITPLANE_MODE_NORM2 | BITPLANE_DIFF;
      }
      else if (DATA_BYTE >= 0x08) {
         bit_shift_inc += 4;
         PARSER_STATE  |= BITPLANE_MODE_NORM6 | BITPLANE_DIFF;
         seq->BPcnt2    = MBcols;
      }
      else {
         bit_shift_inc += 4;
         PARSER_STATE   = VC1_BITPLANE_MODE_RAW;
      }
#ifdef VC1_VIDEO_TRACE
      VC1_TRACE_BEG_BLK
      printf("  <BITPLANE>\n"
             "    invert = %c\n"
             "    Imode  = ",
             ((PARSER_STATE > BITPLANE_INVERT) ? 'y' : 'n'));
      switch (PARSER_STATE & 0x03FF) {
      case VC1_BITPLANE_MODE_RAW:
         printf("Raw\n");
         break;
      case BITPLANE_MODE_COLSKIP:
         printf("Col Skip\n");
         break;
      case BITPLANE_MODE_ROWSKIP:
         printf("Row Skip\n");
         break;
      case BITPLANE_MODE_NORM2:
         printf("Norm-2\n");
         break;
      case BITPLANE_MODE_NORM6:
         printf("Norm-6");
         break;
      case (BITPLANE_DIFF|BITPLANE_MODE_NORM2):
         printf("Diff-2\n");
         break;
      default:
         printf("Diff-6");
      }
      VC1_TRACE_END_BLK
#endif
      seq->BPpos = MBcols * MBrows;
      if (!seq->BPcnt)
         seq->BPcnt = seq->BPpos;
      if (seq->BPpos & 0x0007)
         seq->BPpos += 8;
      memset(bp, 0, (size_t)(seq->BPpos >> 3));
      seq->BPpos = 0;

      if (PARSER_STATE & BITPLANE_MODE_NORM6) {
         if ( (MBcols % 3)&&(!(MBrows % 3)) ) {
            if (MBcols & 0x0001) {
               PARSER_STATE |= BITPLANE_MODE_NORM6_CS1;
               seq->BPcnt   -= MBrows;
               seq->BPpos    = 1;
            }
#ifdef VC1_VIDEO_TRACE
            VC1_TRACE_BEG_BLK
            printf(" 2x3");
            VC1_TRACE_END_BLK
#endif
         }
         else {
            switch (MBcols % 3) {
            case 1:
               PARSER_STATE |= BITPLANE_MODE_NORM6_CS1;
               seq->BPcnt   -= MBrows;
               seq->BPpos    = 1;
               break;
            case 2:
               PARSER_STATE |= BITPLANE_MODE_NORM6_CS2;
               seq->BPcnt   -= MBrows << 1;
               seq->BPpos    = 2;
            }
            if (MBrows & 0x0001) {
               PARSER_STATE |= BITPLANE_MODE_NORM6_RS1;
               seq->BPcnt   -= MBcols;
               seq->BPpos   += MBcols;
            }
            PARSER_STATE |= BITPLANE_MODE_NORM6_32;
#ifdef VC1_VIDEO_TRACE
            VC1_TRACE_BEG_BLK
            printf(" 3x2");
            VC1_TRACE_END_BLK
#endif
         }
         seq->BPcnt /= 6;
#ifdef VC1_VIDEO_TRACE
         VC1_TRACE_BEG_BLK
         printf(" colskip=%hu rowskip=%hu\n",
                ((PARSER_STATE >> 5)&0x0003),
                ((PARSER_STATE >> 7)&0x0001));
         VC1_TRACE_END_BLK
#endif
      }
#if (VC1_VIDEO_TRACE > 1)
      VC1_TRACE_BEG_BLK
      printf("    MBcols = %hu  MBrows = %hu\n"
             "    BPcnt  = %hu  BPpos  = %hu\n",
             MBcols, MBrows,
             seq->BPcnt, seq->BPpos);
      VC1_TRACE_END_BLK
#endif
      ALIGN_DATA_BYTE(PARSER_STATE, bit_shift_inc);
   }

   switch (PARSER_STATE & 0x001F) {
   case VC1_BITPLANE_MODE_RAW:
      seq->payload_len -= (u32)(DATA_PTR - seq->payload);
      seq->payload      = DATA_PTR;
      seq->bit_shift    = BIT_SHIFT_BYTE;
      seq->data         = DATA_BYTE;
      return 1;         /* data is in MACROBLOCK_LAYER */

   case BITPLANE_MODE_COLSKIP:
      if (PARSER_STATE > 0x8000)
         goto BITPLANE_MODE_COLSKIP_1;
      while (seq->BPcnt) {
         GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
         VC1_TRACE_BEG_BLK
         printf("       %2hu> 0x%02hX   1\n",
                seq->BPpos, (u16)DATA_BYTE);
         VC1_TRACE_END_BLK
#endif
         if (DATA_BYTE >= 0x80) {
            PARSER_STATE |= 0x8000;
            seq->BPcnt2   = MBrows;
         }
         else
            seq->BPpos++;
         seq->BPcnt--;
         ALIGN_DATA_BYTE(PARSER_STATE, 1);

         if (PARSER_STATE > 0x8000) {
BITPLANE_MODE_COLSKIP_1:
            while (seq->BPcnt2) {
               GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
               VC1_TRACE_BEG_BLK
               printf("    %2hu,%2hu> 0x%02hX  (",
                      (seq->BPpos % MBcols), (seq->BPpos / MBcols),
                      (u16)DATA_BYTE);
               VC1_TRACE_END_BLK
#endif
               if (seq->BPcnt2 <= 8) {
                  bit_shift_inc = (u8)seq->BPcnt2;
                  seq->BPcnt2   = 0;
                  while (DATA_BYTE) {
                     if (DATA_BYTE >= 0x80)
                        bp_start[seq->BPpos>>3] |= 0x80>>(seq->BPpos&0x0007);
                     BIT_SHIFT_BYTE++;
                     DATA_BYTE <<= 1;
                     if (--bit_shift_inc == 0)
                        break;
                     seq->BPpos += MBcols;
                  }
               }
               else {
                  bit_shift_inc = 0;
                  seq->BPcnt2  -= 8;
                  do {
                     if (DATA_BYTE >= 0x80)
                        bp_start[seq->BPpos>>3] |= 0x80>>(seq->BPpos&0x0007);
                     seq->BPpos += MBcols;
                     DATA_BYTE <<= 1;
                  } while (++bit_shift_inc < 8);
               }
#if (VC1_VIDEO_TRACE > 2)
               VC1_TRACE_BEG_BLK
               printf("%hu)\n", (u16)bit_shift_inc);
               VC1_TRACE_END_BLK
#endif
               ALIGN_DATA_BYTE(PARSER_STATE, bit_shift_inc);
            }
            PARSER_STATE &= 0x7FFF;
            seq->BPpos   %= MBcols;
            seq->BPpos++;
         }
      }
      break;

   case BITPLANE_MODE_ROWSKIP:
      if (PARSER_STATE > 0x8000) {
         bp += seq->BPpos >> 3;
         goto BITPLANE_MODE_ROWSKIP_1;
      }
      while (seq->BPcnt) {
         GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
         VC1_TRACE_BEG_BLK
         printf("       %2hu> 0x%02hX   1\n",
                (seq->BPpos / MBcols), (u16)DATA_BYTE);
         VC1_TRACE_END_BLK
#endif
         if (DATA_BYTE >= 0x80) {
            PARSER_STATE |= 0x8000;
            seq->BPcnt2   = MBcols;
         }
         else {
            seq->BPpos   += MBcols;
            bp            = bp_start + (seq->BPpos >> 3);
         }
         seq->BPcnt--;
         ALIGN_DATA_BYTE(PARSER_STATE, 1);

         if (PARSER_STATE > 0x8000) {
BITPLANE_MODE_ROWSKIP_1:
            while (seq->BPcnt2) {
               GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
               VC1_TRACE_BEG_BLK
               printf("    %2hu,%2hu> 0x%02hX  (",
                      (seq->BPpos % MBcols), (seq->BPpos / MBcols),
                      (u16)DATA_BYTE);
               VC1_TRACE_END_BLK
#endif
               if (seq->BPpos & 0x0007) {
                  tmp1          = DATA_BYTE >> (seq->BPpos & 0x0007);
                  bit_shift_inc = 8 - (u8)(seq->BPpos & 0x0007);
                  if (bit_shift_inc > seq->BPcnt2) {
                     tmp1         &= 0xFF << (bit_shift_inc - seq->BPcnt2);
                     bit_shift_inc = (u8)seq->BPcnt2;
                  }
                  *(bp++) |= tmp1;
               }
               else {
                  if (8 > seq->BPcnt2) {
                     bit_shift_inc = (u8)seq->BPcnt2;
                     tmp1          = 0xFF << (8 - seq->BPcnt2);
                     *bp          |= DATA_BYTE & tmp1;
                  }
                  else {
                     *(bp++)       = DATA_BYTE;
                     bit_shift_inc = 8;
                  }
               }
#if (VC1_VIDEO_TRACE > 2)
               VC1_TRACE_BEG_BLK
               printf("%hu)\n", (u16)bit_shift_inc);
               VC1_TRACE_END_BLK
#endif
               seq->BPpos  += (u16)bit_shift_inc;
               seq->BPcnt2 -= (u16)bit_shift_inc;
               ALIGN_DATA_BYTE(PARSER_STATE, bit_shift_inc);
            }
            PARSER_STATE &= 0x7FFF;
         }
      }
      break;

   case BITPLANE_MODE_NORM2:
      if ( (!seq->BPpos)&&(seq->BPcnt & 0x0001) ) {
         seq->BPpos++;
         GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
         VC1_TRACE_BEG_BLK
         printf("     0, 0> 0x%02hX  ( ,1)\n", (u16)DATA_BYTE);
         VC1_TRACE_END_BLK
#endif
         if (DATA_BYTE >= 0x80)
            *bp = 0x80;
         ALIGN_DATA_BYTE(PARSER_STATE, 1);
      }
      else
         bp += seq->BPpos >> 3;

      while (seq->BPpos < seq->BPcnt) {
         GET_LSB;
         bit_shift_inc = 8;
         do {
            if (DATA_BYTE <= 0x7F)
               tmp1 = 0x01;
            else if (DATA_BYTE <= 0x9F)
               tmp1 = 0x83;
            else if (DATA_BYTE <= 0xBF)
               tmp1 = 0x43;
            else
               tmp1 = 0xC2;
            if (bit_shift_inc < (tmp1 & 0x0F))
               break;
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("    %2hu,%2hu> 0x%02hX  (%hu,%hu)\n",
                   (seq->BPpos % MBcols), (seq->BPpos / MBcols),
                   (u16)DATA_BYTE,
                   (u16)(tmp1 >> 6), (u16)(tmp1 & 0x0F));
            VC1_TRACE_END_BLK
#endif
            BIT_SHIFT_BYTE += (tmp1 & 0x0F);
            DATA_BYTE     <<= (tmp1 & 0x0F);
            bit_shift_inc  -= (tmp1 & 0x0F);
            if (tmp1 > 0x40) {
               if ((seq->BPpos & 7) < 7) {
                  *bp |= (tmp1 & 0xC0) >> (seq->BPpos & 7);
                  if ((seq->BPpos & 7) == 6)
                     bp++;
               }
               else {
                  *(bp++) |= (tmp1 & 0x80) ? 0x01 : 0;
                  *bp     |= (tmp1 & 0x40) ? 0x80 : 0;
               }
            }
            else if ((seq->BPpos & 6) == 6)
               bp++;
            seq->BPpos += 2;
         } while ( (bit_shift_inc)&&(seq->BPpos < seq->BPcnt) );
         ALIGN_DATA_BYTE(PARSER_STATE, 0);
      }
      break;

   case BITPLANE_MODE_NORM6:
      if (PARSER_STATE > 0x8000)
         goto BITPLANE_MODE_NORM6_1;
      while (seq->BPcnt) {
         GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
         VC1_TRACE_BEG_BLK
         printf("    %2hu,%2hu> 0x%02hX  ",
                (seq->BPpos % MBcols), (seq->BPpos / MBcols),
                (u16)DATA_BYTE);
         VC1_TRACE_END_BLK
#endif
         if (DATA_BYTE >= 0x80) {
            bit_shift_inc = 1;
            tmp1          = 0;
         }
         else if (DATA_BYTE >= 0x20) {
            bit_shift_inc = 4;
            tmp1          = (u8)(0x0200 >> (DATA_BYTE >> 4));
         }
         else if (DATA_BYTE >= 0x1C) {
            bit_shift_inc = 6;
            tmp1          = 0xFC;
         }
         else if (DATA_BYTE < 0x10) {
            bit_shift_inc = 8;
            if (DATA_BYTE >= 0x0A)
               tmp1 = 0x04 | (u8)(0x00020000 >> DATA_BYTE);
            else if (DATA_BYTE >= 0x06)
               tmp1 = 0x08 | (u8)(    0x2000 >> DATA_BYTE);
            else if (DATA_BYTE >= 0x03)
               tmp1 = 0x10 | (u8)(    0x0400 >> DATA_BYTE);
            else if (DATA_BYTE)
               tmp1 = 0x20 | (u8)(    0x0100 >> DATA_BYTE);
            else
               tmp1 = 0x40 | 0x80;
         }
         else {
            PARSER_STATE |= (u16)DATA_BYTE << 11;
            ALIGN_DATA_BYTE(PARSER_STATE, 5);
BITPLANE_MODE_NORM6_1:
            GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("(   5)  0x%hu%02hX  ", (PARSER_STATE>>14), (u16)DATA_BYTE);
            VC1_TRACE_END_BLK
#endif
            if (PARSER_STATE > 0xC000) {
               if (DATA_BYTE >= 0x20) {
                  bit_shift_inc = 4;
                  if (DATA_BYTE >= 0x70)
                     tmp1 = 0xF8;
                  else if (DATA_BYTE >= 0x60)
                     tmp1 = 0xF4;
                  else if (DATA_BYTE >= 0x50)
                     tmp1 = 0xEC;
                  else if (DATA_BYTE >= 0x40)
                     tmp1 = 0xDC;
                  else if (DATA_BYTE >= 0x30)
                     tmp1 = 0xBC;
                  else
                     tmp1 = 0x7C;
               }
               else {
                  bit_shift_inc = 8;
                  if (DATA_BYTE == 0x0E)
                     tmp1 = 0xF0;
                  else if (DATA_BYTE == 0x0D)
                     tmp1 = 0xE8;
                  else if (DATA_BYTE == 0x0C)
                     tmp1 = 0xD8;
                  else if (DATA_BYTE == 0x0B)
                     tmp1 = 0xB8;
                  else if (DATA_BYTE == 0x0A)
                     tmp1 = 0x78;
                  else if (DATA_BYTE == 0x09)
                     tmp1 = 0xE4;
                  else if (DATA_BYTE == 0x08)
                     tmp1 = 0xD4;
                  else if (DATA_BYTE == 0x07)
                     tmp1 = 0xB4;
                  else if (DATA_BYTE == 0x06)
                     tmp1 = 0x74;
                  else if (DATA_BYTE == 0x05)
                     tmp1 = 0xCC;
                  else if (DATA_BYTE == 0x04)
                     tmp1 = 0xAC;
                  else if (DATA_BYTE == 0x03)
                     tmp1 = 0x6C;
                  else if (DATA_BYTE == 0x02)
                     tmp1 = 0x9C;
                  else if (DATA_BYTE == 0x01)
                     tmp1 = 0x5C;
                  else
                     tmp1 = 0x3C;
               }
            }
            else {
               bit_shift_inc = 5;
               if (DATA_BYTE >= 0xE0)
                  tmp1 = 0x38;
               else if (DATA_BYTE >= 0xD0)
                  tmp1 = 0x58;
               else if (DATA_BYTE >= 0xC8)
                  tmp1 = 0x98;
               else if (DATA_BYTE >= 0xC0)
                  tmp1 = 0x1C;
               else if (DATA_BYTE >= 0xB0)
                  tmp1 = 0x68;
               else if (DATA_BYTE >= 0xA8)
                  tmp1 = 0xA8;
               else if (DATA_BYTE >= 0xA0)
                  tmp1 = 0x2C;
               else if (DATA_BYTE >= 0x98)
                  tmp1 = 0xC8;
               else if (DATA_BYTE >= 0x90)
                  tmp1 = 0x4C;
               else if (DATA_BYTE >= 0x88)
                  tmp1 = 0x8C;
               else if (DATA_BYTE >= 0x70)
                  tmp1 = 0x70;
               else if (DATA_BYTE >= 0x68)
                  tmp1 = 0xB0;
               else if (DATA_BYTE >= 0x60)
                  tmp1 = 0x34;
               else if (DATA_BYTE >= 0x58)
                  tmp1 = 0xD0;
               else if (DATA_BYTE >= 0x50)
                  tmp1 = 0x54;
               else if (DATA_BYTE >= 0x48)
                  tmp1 = 0x94;
               else if (DATA_BYTE >= 0x38)
                  tmp1 = 0xE0;
               else if (DATA_BYTE >= 0x30)
                  tmp1 = 0x64;
               else if (DATA_BYTE >= 0x28)
                  tmp1 = 0xA4;
               else
                  tmp1 = 0xC4;
            }
            PARSER_STATE &= 0x3FFF;
         }
#if (VC1_VIDEO_TRACE > 2)
         VC1_TRACE_BEG_BLK
         printf("(%02hX,%hu)\n", (u16)tmp1, (u16)bit_shift_inc);
         VC1_TRACE_END_BLK
#endif
         if (PARSER_STATE & BITPLANE_MODE_NORM6_32) {
            /* 3x2 */
            if (tmp1) {
               bp = bp_start + (seq->BPpos >> 3);
               if ((seq->BPpos & 7) == 7) {
                  *bp        |=  tmp1 >> 7;
                  *(bp+1)    |= (tmp1 << 1) & 0xC0;
                  seq->BPpos += MBcols;
                  bp          = bp_start + (seq->BPpos >> 3);
                  *bp        |= (tmp1 >> 4) & 0x01;
                  *(bp+1)    |= (tmp1 << 4) & 0xC0;
               }
               else if ((seq->BPpos & 7) == 6) {
                  *bp        |=  tmp1 >> 6;
                  *(bp+1)    |= (tmp1 << 2) & 0x80;
                  seq->BPpos += MBcols;
                  bp          = bp_start + (seq->BPpos >> 3);
                  *bp        |= (tmp1 >> 3) & 0x03;
                  *(bp+1)    |= (tmp1 << 5) & 0x80;
               }
               else {
                  *bp        |= (tmp1 & 0xE0) >> (seq->BPpos & 7);
                  tmp1      <<= 3;
                  seq->BPpos += MBcols;
                  bp          = bp_start + (seq->BPpos >> 3);
                  *bp        |= (tmp1 & 0xE0) >> (seq->BPpos & 7);
               }
               seq->BPpos -= MBcols - 3;
            }
            else
               seq->BPpos += 3;
            if (seq->BPpos >= seq->BPcnt2) {
               seq->BPpos  += MBcols;
               seq->BPcnt2 += MBcols + MBcols;
            }
         }
         else {
            /* 2x3 */
            if (tmp1) {
               bp = bp_start + (seq->BPpos >> 3);
               if ((seq->BPpos & 7) == 7) {
                  *bp        |=  tmp1 >> 7;
                  *(bp+1)    |= (tmp1 << 1) & 0x80;
                  seq->BPpos += MBcols;
                  bp          = bp_start + (seq->BPpos >> 3);
                  *bp        |= (tmp1 >> 5) & 0x01;
                  *(bp+1)    |= (tmp1 << 3) & 0x80;
                  seq->BPpos += MBcols;
                  bp          = bp_start + (seq->BPpos >> 3);
                  *bp        |= (tmp1 >> 3) & 0x01;
                  *(bp+1)    |= (tmp1 << 5) & 0x80;
               }
               else {
                  *bp        |= (tmp1 & 0xC0) >> (seq->BPpos & 7);
                  tmp1      <<= 2;
                  seq->BPpos += MBcols;
                  bp          = bp_start + (seq->BPpos >> 3);
                  *bp        |= (tmp1 & 0xC0) >> (seq->BPpos & 7);
                  tmp1      <<= 2;
                  seq->BPpos += MBcols;
                  bp          = bp_start + (seq->BPpos >> 3);
                  *bp        |= (tmp1 & 0xC0) >> (seq->BPpos & 7);
               }
               seq->BPpos -= MBcols + MBcols - 2;
            }
            else
               seq->BPpos += 2;
            if (seq->BPpos >= seq->BPcnt2) {
               seq->BPpos  +=  MBcols << 1;
               seq->BPcnt2 += (MBcols << 1) + MBcols;
            }
         }
         seq->BPcnt--;
         ALIGN_DATA_BYTE(PARSER_STATE, bit_shift_inc);
      }

      if (PARSER_STATE & (BITPLANE_MODE_NORM6_CS1|BITPLANE_MODE_NORM6_CS2)) {
         if (PARSER_STATE & BITPLANE_MODE_NORM6_RS1)
            seq->BPcnt = (PARSER_STATE >> 5) & 0x0003;
         PARSER_STATE |= 0x0001;
         seq->BPpos    = 0;
      }
      else if (PARSER_STATE & BITPLANE_MODE_NORM6_RS1) {
         PARSER_STATE |= 0x0002;
         seq->BPpos    = 0;
         bp            = bp_start;
         goto BITPLANE_MODE_NORM6_RS_0;
      }
      break;

   case (BITPLANE_MODE_NORM6 | 0x0001):
      /* colskip */
      if (PARSER_STATE > 0x8000)
         goto BITPLANE_MODE_NORM6_CS_1;
      while (PARSER_STATE & (BITPLANE_MODE_NORM6_CS1|BITPLANE_MODE_NORM6_CS2)) {
         GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
         VC1_TRACE_BEG_BLK
         printf("       %2hu> 0x%02hX   1\n",
                seq->BPpos, (u16)DATA_BYTE);
         VC1_TRACE_END_BLK
#endif
         if (DATA_BYTE >= 0x80) {
            PARSER_STATE |= 0x8000;
            seq->BPcnt2   = MBrows;
         }
         else
            seq->BPpos++;
         PARSER_STATE -= BITPLANE_MODE_NORM6_CS1;
         ALIGN_DATA_BYTE(PARSER_STATE, 1);

         if (PARSER_STATE > 0x8000) {
BITPLANE_MODE_NORM6_CS_1:
            while (seq->BPcnt2) {
               GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
               VC1_TRACE_BEG_BLK
               printf("    %2hu,%2hu> 0x%02hX  (",
                      (seq->BPpos % MBcols), (seq->BPpos / MBcols),
                      (u16)DATA_BYTE);
               VC1_TRACE_END_BLK
#endif
               if (seq->BPcnt2 <= 8) {
                  bit_shift_inc = (u8)seq->BPcnt2;
                  seq->BPcnt2   = 0;
                  while (DATA_BYTE) {
                     if (DATA_BYTE >= 0x80)
                        bp_start[seq->BPpos>>3] |= 0x80>>(seq->BPpos&0x0007);
                     BIT_SHIFT_BYTE++;
                     DATA_BYTE <<= 1;
                     if (--bit_shift_inc == 0)
                        break;
                     seq->BPpos += MBcols;
                  }
               }
               else {
                  bit_shift_inc = 0;
                  seq->BPcnt2  -= 8;
                  do {
                     if (DATA_BYTE >= 0x80)
                        bp_start[seq->BPpos>>3] |= 0x80>>(seq->BPpos&0x0007);
                     seq->BPpos += MBcols;
                     DATA_BYTE <<= 1;
                  } while (++bit_shift_inc < 8);
               }
#if (VC1_VIDEO_TRACE > 2)
               VC1_TRACE_BEG_BLK
               printf("%hu)\n", (u16)bit_shift_inc);
               VC1_TRACE_END_BLK
#endif
               ALIGN_DATA_BYTE(PARSER_STATE, bit_shift_inc);
            }
            PARSER_STATE &= 0x7FFF;
            seq->BPpos   %= MBcols;
            seq->BPpos++;
         }
      }

      if (!seq->BPcnt)
         break;
      PARSER_STATE++;
      seq->BPpos = seq->BPcnt;
      bp         = bp_start;
      goto BITPLANE_MODE_NORM6_RS_0;

   case (BITPLANE_MODE_NORM6 | 0x0002):
      /* rowskip */
      bp += seq->BPpos >> 3;
      goto BITPLANE_MODE_NORM6_RS_1;

BITPLANE_MODE_NORM6_RS_0:
      GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
      VC1_TRACE_BEG_BLK
      printf("       %2hu> 0x%02hX   1\n",
             (seq->BPpos / MBcols), (u16)DATA_BYTE);
      VC1_TRACE_END_BLK
#endif
      if (DATA_BYTE >= 0x80)
         seq->BPcnt2   = MBcols - seq->BPcnt;
      else
         PARSER_STATE += 0x0002; /* invalid PARSER_STATE value */
      ALIGN_DATA_BYTE(PARSER_STATE, 1);

      if (PARSER_STATE & 0x0002) {
BITPLANE_MODE_NORM6_RS_1:
         while (seq->BPcnt2) {
            GET_LSB;
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("    %2hu,%2hu> 0x%02hX  (",
                   (seq->BPpos % MBcols), (seq->BPpos / MBcols),
                   (u16)DATA_BYTE);
            VC1_TRACE_END_BLK
#endif
            if (seq->BPpos & 0x0007) {
               tmp1          = DATA_BYTE >> (seq->BPpos & 0x0007);
               bit_shift_inc = 8 - (u8)(seq->BPpos & 0x0007);
               if (bit_shift_inc > seq->BPcnt2) {
                  tmp1         &= 0xFF << (bit_shift_inc - seq->BPcnt2);
                  bit_shift_inc = (u8)seq->BPcnt2;
               }
               *(bp++) |= tmp1;
            }
            else {
               if (8 > seq->BPcnt2) {
                  bit_shift_inc = (u8)seq->BPcnt2;
                  tmp1          = 0xFF << (8 - seq->BPcnt2);
                  *bp          |= DATA_BYTE & tmp1;
               }
               else {
                  *(bp++)       = DATA_BYTE;
                  bit_shift_inc = 8;
               }
            }
#if (VC1_VIDEO_TRACE > 2)
            VC1_TRACE_BEG_BLK
            printf("%hu)\n", (u16)bit_shift_inc);
            VC1_TRACE_END_BLK
#endif
            seq->BPpos  += (u16)bit_shift_inc;
            seq->BPcnt2 -= (u16)bit_shift_inc;
            ALIGN_DATA_BYTE(PARSER_STATE, bit_shift_inc);
         }
      }
   }

   seq->payload_len -= (u32)(DATA_PTR - seq->payload);
   seq->payload      = DATA_PTR;
   seq->bit_shift    = BIT_SHIFT_BYTE;
   seq->data         = DATA_BYTE;

   bp = bp_start;
   if (PARSER_STATE & BITPLANE_DIFF) {
      /* diff operation */
      data          =
      bit_shift_inc = (PARSER_STATE & BITPLANE_INVERT) ? 0xFF : 0x00;
      bit_shift     =
      tmp1          = 0x80;
      ptr           = bp;
      for (seq->BPcnt = 0; seq->BPcnt < MBrows; seq->BPcnt++) {
         for (seq->BPcnt2 = 0; seq->BPcnt2 < MBcols; seq->BPcnt2++) {
            *bp          ^= bit_shift_inc & tmp1;
            bit_shift_inc = (*bp & tmp1) ? 0xFF : 0x00;
            if (tmp1 == 0x01) {
               bp++;
               tmp1   = 0x80;
            }
            else
               tmp1 >>= 1;
            if (seq->BPcnt) {
               if (bit_shift == 0x01) {
                  ptr++;
                  bit_shift   = 0x80;
               }
               else
                  bit_shift >>= 1;
               if ( ( (bit_shift_inc) &&(!(*ptr & bit_shift)) )||
                    ( (!bit_shift_inc)&&  (*ptr & bit_shift)  ) )
                  bit_shift_inc = data;
            }
         }
         bit_shift_inc = (*ptr & bit_shift) ? 0xFF : 0x00;
      }
#if (VC1_VIDEO_TRACE > 2)
      VC1_TRACE_BEG_BLK
      if (data)
         printf("    DIFF  INVERT\n");
      else
         printf("    DIFF\n");
      VC1_TRACE_END_BLK
#endif
   }
   else if (PARSER_STATE & BITPLANE_INVERT) {
      /* invert bitplane */
      seq->BPcnt = MBcols * MBrows;
      if (seq->BPcnt & 0x0007)
         seq->BPcnt += 8;
      seq->BPcnt >>= 3;
      do {
         *(bp++) ^= 0xFF;
      } while (--(seq->BPcnt));
#if (VC1_VIDEO_TRACE > 2)
      VC1_TRACE_BEG_BLK
      printf("    INVERT\n");
      VC1_TRACE_END_BLK
#endif
   }

#if (VC1_VIDEO_TRACE > 1)
   VC1_TRACE_BEG_BLK
   bp   = bp_start;
   tmp1 = 0x80;
   for (seq->BPcnt = 0; seq->BPcnt < MBrows; seq->BPcnt++) {
      printf("    ");
      for (seq->BPcnt2 = 0; seq->BPcnt2 < MBcols; seq->BPcnt2++) {
         printf("%c", ((*bp & tmp1) ? '1' : '0'));
         if (tmp1 == 0x01) {
            bp++;
            tmp1   = 0x80;
         }
         else
            tmp1 >>= 1;
      }
      printf("\n");
   }
   VC1_TRACE_END_BLK
#endif
   return 1;

VC1_VIDEO_read_BITPLANE_END:
   seq->bit_shift   = BIT_SHIFT_BYTE;
   seq->data        = DATA_BYTE;
   seq->payload_len = 0;
   return 0;

#undef BITPLANE_MODE_COLSKIP
#undef BITPLANE_MODE_ROWSKIP
#undef BITPLANE_MODE_NORM2
#undef BITPLANE_MODE_NORM6
#undef BITPLANE_MODE_NORM6_CS1
#undef BITPLANE_MODE_NORM6_CS2
#undef BITPLANE_MODE_NORM6_RS1
#undef BITPLANE_MODE_NORM6_32
#undef BITPLANE_DIFF
#undef BITPLANE_INVERT
}

