/***************************************************************************
                          ListDLG.c  -  description
                             -------------------
   Stuff needed for handling the list dialog.
                             -------------------
    begin                : Sun Feb 27 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/*#define SHOW_ALL_PIDS   1*/
#include <stdio.h>
#ifndef WIN32_LEAN_AND_MEAN
# define WIN32_LEAN_AND_MEAN
#endif
#include <windows.h>
#include <commctrl.h>

#include "winmain.h"
#include "resource.h"
#include "ListDLG.h"
#include "DLU.h"
#include "MPEG2_SYS.h"
#include "PMT_SECTION.h"
#include "mpeg_es.h"
#include "mpeg_ts.h"
#include "mpeg_ps.h"

#define            ListDLG_cols             2
static int         ListDLG_ColWid[]       = {(int)DLUX(40),
                                             (int)DLUX(80)};
static const char *SelectProgram_Title    = "Select Program",
                  *SelectProgram_ColTxt[] = {"Program",
                                             "PMT PID"},
                  *SelectStreams_Title    = "Select Streams",
                  *SelectStreams_ColTxt[] = {"PID",
                                             "Type"};
static FILE_PTR    cur_fp;
static int         TS;

static int  ListDLG_notify(LPNMHDR hdr);
static HWND ListDLG_init(HWND        hw,
                         const char *title,
                         const char *col_txt[],
                         int         col_wid[],
                         int         num_cols);

static BOOL CALLBACK SelectProgram_DLGproc(HWND hw, UINT msg,
                                           WPARAM p1, LPARAM p2);
static void          SelectProgram_init(HWND list);
static void          SelectProgram_choice(HWND list, int idx);

static BOOL CALLBACK SelectStreams_DLGproc(HWND hw, UINT msg,
                                           WPARAM p1, LPARAM p2);
static void          SelectStreams_init(HWND list);
static void          SelectStreams_choice(HWND list, int idx);
static void          SelectStreams_create_prog(HWND hw);


void ListDLG(FILE_PTR fp, int select_program, int TS_flag)
{
   cur_fp = fp;
   TS     = TS_flag;
   if (select_program) {
      DialogBox(app, MAKEINTRESOURCE(IDD_LIST_DIALOG), 
                win, SelectProgram_DLGproc);
   }
   else {
      DialogBox(app, MAKEINTRESOURCE(IDD_LIST_DIALOG), 
                win, SelectStreams_DLGproc);
   }
   cur_fp = FILE_PTR_BAD_VALUE;
}


static int ListDLG_notify(LPNMHDR hdr)
{
   int i = -1;

   if (hdr->idFrom == IDC_LIST_DIALOG_LIST) {
      switch (hdr->code) {
      case NM_CLICK:
         goto FIND_SELECTED_ITEM;

      case LVN_KEYDOWN:
      {
         LV_KEYDOWN *kd = (LV_KEYDOWN FAR *)hdr;

         if (kd->wVKey == VK_SPACE) {
FIND_SELECTED_ITEM:
            i = ListView_GetNextItem(hdr->hwndFrom, -1, LVNI_FOCUSED);
         }
      }
      }
   }
   return i;
}

static HWND ListDLG_init(HWND        hw,
                         const char *title,
                         const char *col_txt[],
                         int         col_wid[],
                         int         num_cols)
{
   HWND       list;
   HIMAGELIST img_list;
   HBITMAP    img;
   LV_COLUMN  col_hdr;
   int        i;


   /* Set title bar text */
   SetWindowText(hw, title);

   list = GetDlgItem(hw, IDC_LIST_DIALOG_LIST);
   if (!list)
      return NULL;

   /* Initialize image list */
   img_list = ImageList_Create(16, 16, TRUE, 2, 1);
   img      = LoadBitmap(app, MAKEINTRESOURCE(IDB_UNCHECKED_BITMAP));
   ImageList_AddMasked(img_list, img, RGB(255, 255, 255));
   DeleteObject(img);
   img      = LoadBitmap(app, MAKEINTRESOURCE(IDB_CHECKED_BITMAP));
   ImageList_AddMasked(img_list, img, RGB(255, 255, 255));
   DeleteObject(img);
   ListView_SetImageList(list, img_list, LVSIL_SMALL);

   /* Set column headers */
   col_hdr.mask       = LVCF_FMT | LVCF_TEXT | LVCF_WIDTH | LVCF_SUBITEM;
   col_hdr.fmt        = LVCFMT_LEFT;
   col_hdr.cchTextMax = 0;
   for (i = 0; i < num_cols; i++) {
      col_hdr.iSubItem = i;
      col_hdr.cx       = col_wid[i];
      col_hdr.pszText  = (LPTSTR)col_txt[i];
      if (ListView_InsertColumn(list, i, &col_hdr) < 0)
         break;
   }

   return list;
}


static BOOL CALLBACK SelectProgram_DLGproc(HWND hw, UINT msg,
                                           WPARAM p1, LPARAM p2)
{
   switch (msg) {
   case WM_NOTIFY:   /* controls are signalling an event */
   {
      int     idx;
      LPNMHDR hdr = (LPNMHDR)p2;

      idx = ListDLG_notify(hdr);
      if (idx >= 0)
         SelectProgram_choice(hdr->hwndFrom, idx);
      break;
   }

   case WM_COMMAND:
      if ( (HIWORD(p1) == BN_CLICKED)&&(LOWORD(p1) == IDOK) )
         EndDialog(hw, 0);
      break;

   case WM_INITDIALOG:
   {
      HWND list = ListDLG_init(hw,
                               SelectProgram_Title,
                               SelectProgram_ColTxt,
                               ListDLG_ColWid,
                               ListDLG_cols);
      if (list)
         SelectProgram_init(list);
      return TRUE;
   }
   }
   return FALSE;
}

static void SelectProgram_init(HWND list)
{
   char    buf[16];
   LV_ITEM item;
   u16     value;

   item.iItem   = 0;   /*row_num;*/
   item.pszText = buf; /*entry text*/
   item.iImage  = 0;   /*icon idx num*/
   while (1) {
      value = sys.get_program(sys.psi_pes[sys.PAT_idx].common.mem,
                              (u16)item.iItem);
      if (!value)
         break;
      sprintf(buf, "%hu", value);
      item.mask     = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
      item.iSubItem = 0;               /*col_num*/
      item.lParam   = (LPARAM)value;   /*program number (user defined)*/
      ListView_InsertItem(list, &item);
      if (value == sys.program_number)
         SelectProgram_choice(list, item.iItem);

      value = sys.get_pid(sys.psi_pes[sys.PAT_idx].common.mem, value);
      sprintf(buf, "%hX", value);
      item.mask     = LVIF_TEXT;
      item.iSubItem = 1;               /*col_num*/
      ListView_SetItem(list, &item);

      item.iItem++;
   }
   
}

static void SelectProgram_choice(HWND list, int idx)
{
   LV_ITEM item;

   item.iItem    = 0;   /*row_num;*/
   item.iSubItem = 0;   /*col_num*/
   while (1) {
      item.mask = LVIF_PARAM;
      if (!ListView_GetItem(list, &item))
         break;
      else if (item.iItem == idx) {
         /* this program was chosen */
         item.iImage        = 1;
         sys.program_number = (u16)item.lParam;
      }
      else {
         /* this program was not chosen */
         item.iImage = 0;
      }

      item.mask |= LVIF_IMAGE;
      ListView_SetItem(list, &item);
      item.iItem++;
   }
}

static BOOL CALLBACK SelectStreams_DLGproc(HWND hw, UINT msg,
                                           WPARAM p1, LPARAM p2)
{
   switch (msg) {
   case WM_NOTIFY:   /* controls are signalling an event */
   {
      int     idx;
      LPNMHDR hdr = (LPNMHDR)p2;

      idx = ListDLG_notify(hdr);
      if (idx >= 0)
         SelectStreams_choice(hdr->hwndFrom, idx);
      break;
   }

   case WM_COMMAND:
      if ( (HIWORD(p1) == BN_CLICKED)&&(LOWORD(p1) == IDOK) ) {
         SelectStreams_create_prog(hw);
         EndDialog(hw, 0);
      }
      break;

   case WM_INITDIALOG:
   {
      HWND list = ListDLG_init(hw,
                               SelectStreams_Title,
                               SelectStreams_ColTxt,
                               ListDLG_ColWid,
                               ListDLG_cols);
      if (list)
         SelectStreams_init(list);
      return TRUE;
   }
   }
   return FALSE;
}

#define NUM_PIDS     16
#define VIDEO_STREAM 0x1000000
#define AUDIO_STREAM 0x2000000

static void SelectStreams_init(HWND list)
{
   char     buf[16];
   u16      PID[NUM_PIDS];
   u8       stream_type[NUM_PIDS];
   LV_ITEM  item;
   FILE_POS fp_pos;
   u32      num, i, j,
            vid = 0,
            aud = 0;

   if (FILE_PTR_BAD(cur_fp))
      return;
   FILE_GETPOS(cur_fp, fp_pos);
   FILE_REWIND(cur_fp);
   if (TS)
      num = MPEG_TS_get_PIDs(cur_fp, PID, stream_type, NUM_PIDS);
   else
      num = MPEG_PS_get_PIDs(cur_fp, PID, stream_type, NUM_PIDS);
   FILE_SETPOS(cur_fp, fp_pos);

   item.iItem  = 0;  /*row_num;*/
   item.iImage = 0;  /*icon idx num*/
   for (i = 0; i < num; i++) {
#ifndef SHOW_ALL_PIDS
      if (!stream_type[i])
         continue;
#endif
      sprintf(buf, "%hX", PID[i]);
      item.mask     = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
      item.pszText  = buf;             /*entry text*/
      item.iSubItem = 0;               /*col_num*/
      item.lParam   = (LPARAM)PID[i];  /*type | pid (user defined)*/
      item.lParam  |= (LPARAM)stream_type[i] << 16;
      switch (stream_type[i]) {
      case STREAM_TYPE_MPEG1_VIDEO:
      case STREAM_TYPE_MPEG2_VIDEO:
         item.lParam |= VIDEO_STREAM;
         vid++;
         break;
      case STREAM_TYPE_MPEG1_AUDIO:
      case STREAM_TYPE_MPEG2_AUDIO:
      case STREAM_TYPE_ATSC_AC3_AUDIO:
         item.lParam |= AUDIO_STREAM;
         aud++;
         break;
      }
      ListView_InsertItem(list, &item);

#ifdef SHOW_ALL_PIDS
      if (stream_type[i])
#endif
      {  /* check if this stream is active */
         for (j = 0; j < MPEG2_SYS_PIDS; j++) {
            if (sys.PID[j] == PID[i]) {
               SelectStreams_choice(list, item.iItem);
               break;
            }
         }
      }

#ifdef SHOW_ALL_PIDS
      if (PID[i] == TS_PKT_PAT_ID)
         item.pszText = "Program Association Table";
      else if (PID[i] == TS_PKT_CAT_ID)
         item.pszText = "Conditional Access Table";
      else if (PID[i] == TS_PKT_ATSC_SI_ID)
         item.pszText = "ATSC System Info & Program Guide Data";
      else {
#endif
         switch (stream_type[i]) {
         case STREAM_TYPE_MPEG1_VIDEO:
            item.pszText = "MPEG1 Video";
            break;
         case STREAM_TYPE_MPEG2_VIDEO:
            item.pszText = "MPEG2 Video";
            break;
         case STREAM_TYPE_MPEG1_AUDIO:
            item.pszText = "MPEG1 Audio";
            break;
         case STREAM_TYPE_MPEG2_AUDIO:
            item.pszText = "MPEG2 Audio";
            break;
         case STREAM_TYPE_PRIVATE:
            item.pszText = "Private";
            break;
         case STREAM_TYPE_PES_PRIVATE:
            item.pszText = "Private PES";
            break;
         case STREAM_TYPE_ISO_13522_MHEG:
            item.pszText = "ISO 13522 MHEG";
            break;
         case STREAM_TYPE_ISO_13818_1_DSMCC:
            item.pszText = "ISO 13818 1 DSMCC";
            break;
         case STREAM_TYPE_ITUT_H2221:
            item.pszText = "ITUT H2221";
            break;
         case STREAM_TYPE_ISO_13818_6_A:
            item.pszText = "ISO 13818 6 A";
            break;
         case STREAM_TYPE_ISO_13818_6_B:
            item.pszText = "ISO 13818 6 B";
            break;
         case STREAM_TYPE_ISO_13818_6_C:
            item.pszText = "ISO 13818 6 C";
            break;
         case STREAM_TYPE_ISO_13818_6_D:
            item.pszText = "ISO 13818 6 D";
            break;
         case STREAM_TYPE_ISO_13818_1_AUX:
            item.pszText = "ISO 13818 1 AUX";
            break;
         case STREAM_TYPE_ATSC_AC3_AUDIO:
            item.pszText = "ATSC AC3 Audio";
            break;
         default:
            if ( (0x80 <= stream_type[i])&&(stream_type[i] <= 0xAF) )
               item.pszText = "ATSC Private Stream";
            else
               item.pszText = "Unknown";
         }
#ifdef SHOW_ALL_PIDS 
      }
#endif
      item.mask     = LVIF_TEXT;
      item.iSubItem = 1;   /*col_num*/
      ListView_SetItem(list, &item);

      item.iItem++;
   }

   /* Add two extra entries, so the user can disable
        the audio or video streams. */
   if (vid) {
      item.mask     = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
      item.pszText  = "1FFF";          /*entry text*/
      item.iSubItem = 0;               /*col_num*/
      item.lParam   = (LPARAM)0x1FFF;  /*type | pid (user defined)*/
      item.lParam  |= VIDEO_STREAM;
      ListView_InsertItem(list, &item);
      item.pszText  = "Disable Video";
      item.mask     = LVIF_TEXT;
      item.iSubItem = 1;   /*col_num*/
      ListView_SetItem(list, &item);
      item.iItem++;
   }

   if (aud) {
      item.mask     = LVIF_TEXT | LVIF_IMAGE | LVIF_PARAM;
      item.pszText  = "1FFF";          /*entry text*/
      item.iSubItem = 0;               /*col_num*/
      item.lParam   = (LPARAM)0x1FFF;  /*type | pid (user defined)*/
      item.lParam  |= AUDIO_STREAM;
      ListView_InsertItem(list, &item);
      item.pszText  = "Disable Audio";
      item.mask     = LVIF_TEXT;
      item.iSubItem = 1;   /*col_num*/
      ListView_SetItem(list, &item);
   }
}

static void SelectStreams_choice(HWND list, int idx)
{
   LV_ITEM item;
   LPARAM  type;

   item.mask     = LVIF_PARAM;
   item.iItem    = idx; /*row_num;*/
   item.iSubItem = 0;   /*col_num*/
   if (!ListView_GetItem(list, &item))
      return;

   type = 0xFF000000 & item.lParam;
   if (!type)
      return;

   item.iItem    = 0;   /*row_num;*/
   item.iSubItem = 0;   /*col_num*/
   while (1) {
      item.mask = LVIF_PARAM;
      if (!ListView_GetItem(list, &item))
         break;
      else if (item.lParam & type) {
         if (item.iItem == idx) {
            /* this stream was chosen */
            item.iImage = 1;
         }
         else {
            /* this stream was not chosen */
            item.iImage = 0;
         }
         item.mask |= LVIF_IMAGE;
         ListView_SetItem(list, &item);
      }
      item.iItem++;
   }
}

static void SelectStreams_create_prog(HWND hw)
{
   LV_ITEM item;
   u16     vid_PID  = TS_PKT_NUL_ID, 
           aud_PID  = TS_PKT_NUL_ID;
   u8      vid_type = 0,
           aud_type = 0;

   hw = GetDlgItem(hw, IDC_LIST_DIALOG_LIST);
   if (!hw)
      return;

   item.mask     = LVIF_PARAM | LVIF_IMAGE;
   item.iItem    = 0;   /*row_num;*/
   item.iSubItem = 0;   /*col_num*/
   while (1) {
      if (!ListView_GetItem(hw, &item))
         break;
      else if (item.iImage) {
         if (item.lParam & VIDEO_STREAM) {
            vid_PID  = (u16)(0x0000FFFF & item.lParam);
            vid_type = (u8)((0x00FF0000 & item.lParam) >> 16);
         }
         else if (item.lParam & AUDIO_STREAM) {
            aud_PID  = (u16)(0x0000FFFF & item.lParam);
            aud_type = (u8)((0x00FF0000 & item.lParam) >> 16);
         }
      }
      item.iItem++;
   }

   if (TS) {
      if ( (vid_PID < TS_PKT_NUL_ID)||(aud_PID < TS_PKT_NUL_ID) ) {
         if ( (sys.add_PAT_entry)&&(sys.add_PMT_entry) ) {
            /* create the TS program */
            sys.program_number = 1;
            sys.add_PAT_entry(sys.psi_pes[sys.PAT_idx].common.mem,
                              sys.program_number, 0x1FFE);
            sys.add_PMT_entry(sys.psi_pes[sys.PMT_idx].common.mem,
                              sys.program_number, TS_PKT_NUL_ID,
                              vid_PID, vid_type,
                              aud_PID, aud_type);
         }
      }
   }
   else {
      int mod;

      if (vid_PID < TS_PKT_NUL_ID) {
         mod      = MPEG_ES_find_module(vid_type);
         vid_type = (mod < 0) ? 0 : ((u8)mod + 1);
         if (!vid_type)
            vid_PID = TS_PKT_NUL_ID;
         else
            vid_type--;
      }
      if (aud_PID < TS_PKT_NUL_ID) {
         mod      = MPEG_ES_find_module(aud_type);
         aud_type = (mod < 0) ? 0 : ((u8)mod + 1);
         if (!aud_type)
            aud_PID = TS_PKT_NUL_ID;
         else
            aud_type--;
      }

      if ( (vid_PID < TS_PKT_NUL_ID)||
           (aud_PID < TS_PKT_NUL_ID) ) {
         MPEG2_SYS_reset_PIDs();
         sys.program_number = 1;
      
         if (vid_PID < TS_PKT_NUL_ID)
            sys.PID[vid_type] = vid_PID;

         if (aud_PID < TS_PKT_NUL_ID)
            sys.PID[aud_type] = aud_PID;
      }
   }
}

#undef NUM_PIDS
#undef VIDEO_STREAM
#undef AUDIO_STREAM

#undef ListDLG_cols
#undef SHOW_ALL_PIDS 

