/***************************************************************************
                       file_actions.c  -  description
                             -------------------
   Functions for handling the files.
                             -------------------
    begin                : Mon Mar 7 2005
    copyright            : (C) 2005 by Dennis Lew
    email                : dennis_L76@yahoo.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <string.h>

#include "commands.h"
#include "TS_PKT.h"
#include "ADAPT_FIELD.h"
#include "PMT_SECTION.h"
#include "MPEG2_SYS.h"
#include "mpeg_es.h"
#include "mpeg_ps.h"
#include "mpeg_ts.h"
#include "common/Timer.h"
#ifdef __DVD_INPUT__
# include "dvd/dvd_input.h"
# include "dvd/mpeg_dvd.h"
#endif
#include "file_actions.h"

typedef u32 (*GET_PIDs)(FILE_PTR, u16 *, u8 *, u32);

static void Display_PIDs(FILE_PTR fp, GET_PIDs list);
#ifdef __DVD_INPUT__
static void Display_DVD_PIDs(DVD_PTR dvd);
static void Display_DVD_info(DVD_PTR dvd);
#endif
#ifdef MPEG2_SYS_CLK64
static void Display_DataRates(FILE_PTR fp);
#endif
static void Display_PAT(FILE_PTR fp);
static void Display_PMT(FILE_PTR fp, u16 prog_num);
static void Display_stream_type(u16 PID, u8 stream_type);
#ifdef _RECORD_STATS_
static void Display_stats();
#endif

void start_ES(FILE_PTR fp, u8 stream_type)
{
   FILE_REWIND(fp);
   sys.data.command = (sys.jump_offset == 0.0) ? CMD_PLAY : CMD_JUMP;
   MPEG_ES_Play(fp, stream_type);
#ifdef _RECORD_STATS_
   Display_stats();
#endif
}

void start_TS(APP_PARAMS *params)
{
   sys.program_number = 0;
   FILE_REWIND(params->fp);
      
   /* carry out request */
   switch (params->req) {
   case DISPLAY_PIDS:
      Display_PIDs(params->fp, MPEG_TS_get_PIDs);
      return;
   case DISPLAY_RATES:
#ifdef MPEG2_SYS_CLK64
      Display_DataRates(params->fp);
#endif
      return;
   case DISPLAY_PAT:
      Display_PAT(params->fp);
      return;
   case DISPLAY_PMT:
      Display_PMT(params->fp, params->prog_num);
      return;
   case PLAY_PROG:
      /* load the PAT, to make sure the program is valid */
      MPEG_TS_load_PAT(params->fp);
      if (sys.get_pid(sys.psi_pes[sys.PAT_idx].common.mem, params->prog_num) < TS_PKT_NUL_ID)
         sys.program_number = params->prog_num;
      else
         printf("Could not find program (%04hX), playing default...\n", params->prog_num);
      break;
   case PLAY_STREAMS:
   {
      u8 vid_type, aud_type;
      
      /* find correct stream type values */
      if (params->video_PID < TS_PKT_NUL_ID) {
         vid_type = MPEG_TS_stream_type(params->fp, params->video_PID);
         if (!vid_type) {
            printf("Cannot play stream (%04hX)\n", params->video_PID);
            params->video_PID = TS_PKT_NUL_ID;
         }
      }
      if (params->audio_PID < TS_PKT_NUL_ID) {
         aud_type = MPEG_TS_stream_type(params->fp, params->audio_PID);
         if (!aud_type) {
            printf("Cannot play stream (%04hX)\n", params->audio_PID);
            params->audio_PID = TS_PKT_NUL_ID;
         }
      }

      sys.program_number = 0;
      if ( (params->video_PID >= TS_PKT_NUL_ID)&&
           (params->audio_PID >= TS_PKT_NUL_ID) )
         printf("No streams to play, playing default...\n");
      else if ( (sys.add_PAT_entry)&&(sys.add_PMT_entry) ) {
         /* create the TS program */
         sys.program_number = 1;
         sys.add_PAT_entry(sys.psi_pes[sys.PAT_idx].common.mem,
                           sys.program_number, 0x1FFE);
         sys.add_PMT_entry(sys.psi_pes[sys.PMT_idx].common.mem,
                           sys.program_number, params->PCR_PID,
                           params->video_PID, vid_type,
                           params->audio_PID, aud_type);
      }
   }
   }

   if (!sys.program_number) {
      MPEG_TS_Find_Program(params->fp);
      if (!sys.program_number)
         MPEG_TS_Create_Program(params->fp);
   }

   if (sys.program_number) {
      sys.data.command = (sys.jump_offset == 0.0) ? CMD_PLAY : CMD_JUMP;
      sys.offset       = 0;
      MPEG2_SYS_Play(params->fp, MPEG_TS_Play);
#ifdef _RECORD_STATS_
      Display_stats();
#endif
   }
   else
      printf("Cannot find anything to play.\n\n");
}

void start_PS(APP_PARAMS *params)
{
   sys.program_number = 0;
#ifdef __DVD_INPUT__
   if (DVD_PTR_OK(params->dvd)) {
      DVD_REWIND(params->dvd);
   }
   else
#endif
      FILE_REWIND(params->fp);
      
   /* carry out request */
   switch (params->req) {
   case DISPLAY_PIDS:
#ifdef __DVD_INPUT__
      if (DVD_PTR_OK(params->dvd))
         Display_DVD_PIDs(params->dvd);
      else
#endif
         Display_PIDs(params->fp, MPEG_PS_get_PIDs);
      return;
   case PLAY_STREAMS:
   {
      int i;
      u8  vid_type, aud_type;

      /* find correct stream type values */
      if (params->video_PID < TS_PKT_NUL_ID) {
#ifdef __DVD_INPUT__
         if (DVD_PTR_OK(params->dvd))
            vid_type = MPEG_DVD_stream_type(params->dvd, params->video_PID);
         else
#endif
            vid_type = MPEG_PS_stream_type(params->fp, params->video_PID);
         i        = MPEG_ES_find_module(vid_type);
         vid_type = (i < 0) ? 0 : ((u8)i + 1);
         if (!vid_type) {
            printf("Cannot play stream (%04hX)\n", params->video_PID);
            params->video_PID = TS_PKT_NUL_ID;
         }
         else
            vid_type--;
      }
      if (params->audio_PID < TS_PKT_NUL_ID) {
#ifdef __DVD_INPUT__
         if (DVD_PTR_OK(params->dvd))
            aud_type = MPEG_DVD_stream_type(params->dvd, params->audio_PID);
         else
#endif
            aud_type = MPEG_PS_stream_type(params->fp, params->audio_PID);
         i        = MPEG_ES_find_module(aud_type);
         aud_type = (i < 0) ? 0 : ((u8)i + 1);
         if (!aud_type) {
            printf("Cannot play stream (%04hX)\n", params->audio_PID);
            params->audio_PID = TS_PKT_NUL_ID;
         }
         else
            aud_type--;
      }

      if ( (params->video_PID >= TS_PKT_NUL_ID)&&
           (params->audio_PID >= TS_PKT_NUL_ID) ) {
         printf("No streams to play, playing default...\n");
         sys.program_number = 0;
      }
      else {
         sys.program_number = 1;
      
         if (params->video_PID < TS_PKT_NUL_ID)
            sys.PID[vid_type] = params->video_PID;

         if (params->audio_PID < TS_PKT_NUL_ID)
            sys.PID[aud_type] = params->audio_PID;
      }
#ifdef __DVD_INPUT__
      break;
#endif
   }
#ifdef __DVD_INPUT__
   case DISPLAY_DVD_INFO:
      Display_DVD_info(params->dvd);
      return;
#endif
   }

   if (!sys.program_number) {
#ifdef __DVD_INPUT__
      if (DVD_PTR_OK(params->dvd))
         MPEG_DVD_Create_Program(params->dvd);
      else
#endif
         MPEG_PS_Create_Program(params->fp);
   }

   if (sys.program_number) {
      sys.data.command = (sys.jump_offset == 0.0) ? CMD_PLAY : CMD_JUMP;
      sys.offset       = 0;
#ifdef __DVD_INPUT__
      if (DVD_PTR_OK(params->dvd)) {
         printf("TITLE: %hu/%hu\n", dvd_title, dvd_title_last);
         MPEG_DVD_Play(params->dvd, params->chapter);
         params->chapter = 0;
      }
      else
#endif
         MPEG2_SYS_Play(params->fp, MPEG_PS_Play);
#ifdef _RECORD_STATS_
      Display_stats();
#endif
   }
   else
      printf("Cannot find anything to play.\n\n");
}


#define NUM_PIDS  16

static void Display_PIDs(FILE_PTR fp, GET_PIDs list)
{
   u16 PID[NUM_PIDS];
   u8  stream_type[NUM_PIDS];
   u32 num, i;

   num = list(fp, PID, stream_type, NUM_PIDS);
   if (!num)
      return;

   printf("    PID  Type\n");
   for (i = 0; i < num; i++) {
      printf("   %04hX  ", PID[i]);
      Display_stream_type(PID[i], stream_type[i]);
   }
   printf("\n");
}

#ifdef __DVD_INPUT__
static void Display_DVD_PIDs(DVD_PTR dvd)
{
   u16 PID[NUM_PIDS];
   u8  stream_type[NUM_PIDS];
   u32 num, i;
   u16 lang;

   num = MPEG_DVD_get_PIDs(dvd, PID, stream_type, NUM_PIDS);
   if (!num)
      return;

   printf("TITLE: %hu/%hu\n", dvd_title, dvd_title_last);
   printf("    PID  LANG  Type\n");
   for (i = 0; i < num; i++) {
      printf("   %04hX  ", PID[i]);
      if ( ((PID[i]&0x00F0) == 0x00E0)||
           ((lang = DVD_AUDIO_LANG(dvd, PID[i])) == 0xFFFF) )
         printf("      ");
      else
         printf(" %c%c   ", (char)(lang>>8), (char)(lang&0x00FF));
      Display_stream_type(PID[i], stream_type[i]);
   }
   printf("\n");
}

static void Display_DVD_info(DVD_PTR dvd)
{
   const char *title;
   u16         chapters[NUM_PIDS];
   u32         num, i;
   
   num = MPEG_DVD_get_info(dvd, &title, chapters, NUM_PIDS);
   if (!num)
      return;
   
   printf("<< %s >>\n"
          "    Title  Chapters\n", title);
   for (i = 0; i < num; i++) {
      printf("    %5u  %8hu\n", (i+1), chapters[i]);
   }
}
#endif

#ifdef MPEG2_SYS_CLK64
static void Display_DataRates(FILE_PTR fp)
{
   const u8       *pkt[MPEG2_SYS_BUF_PKTS+1], *aligned_pkt;
   FILE_BYTE_COUNT i, i2, num, tot_pkts, num_ids, tot_afs,
                   last_pkt[MPEG2_SYS_PIDS], num_PCRs[MPEG2_SYS_PIDS];
   u16             cont_cnt[MPEG2_SYS_PIDS];
   TS_PKT          ts, ts_pkt[MPEG2_SYS_PIDS];
   s64             last_PCR[MPEG2_SYS_PIDS];
   double          min[MPEG2_SYS_PIDS], max[MPEG2_SYS_PIDS], 
                   avg[MPEG2_SYS_PIDS], rate;
   ADAPT_FIELD     af;
   TIME            t1, t2, 
                   tot_t = {0, 0};
   u8             *buf   = sys.buf;

   /* initialize packet IDS */
   for (i = 0; i < MPEG2_SYS_PIDS; i++) {
      ts_pkt[i].PID = 0xFFFF;

      num_PCRs[i] = 0;
      min[i]      = 1.0e20;
      max[i]      = 0.0;
      avg[i]      = 0.0;
   }

   pkt[0]   = NULL;
   tot_pkts = num_ids = tot_afs = 0;
   while (1) {
      FILE_READ(fp, buf, (TS_PKT_SIZE*MPEG2_SYS_BUF_PKTS), num);
      num /= TS_PKT_SIZE;
      if (num < 1)
         break;
      TIMER_curtime(&t1);

      for (i = 0, aligned_pkt = buf; i < num; i++, aligned_pkt += TS_PKT_SIZE) {
         if ( (!pkt[i])||(*(pkt[i]) != TS_PKT_SYNC_BYTE) ) {
            pkt[i] = TS_PKT_find(aligned_pkt, (u32)(TS_PKT_SIZE*(num-i)));
            if (!pkt[i])
               break;   /* this and all following packets read into buf are bad */
         }
         pkt[i]++;   /* skip the TS_SYNC_BYTE */

         if ( (i == (MPEG2_SYS_BUF_PKTS-1))&&(pkt[i] > (aligned_pkt+1)) ) {
            tot_pkts += num - 1;

            /* the tail end of this packet has not been read yet, so
                 lets do it now (re-fill entire buffer) */
            num = TS_PKT_SIZE + aligned_pkt - pkt[i];
            memmove((buf+1), pkt[i], num);   /* move this packet into the first
                                                  packet position */
            pkt[0] = buf + 1;

            num++;   /* add the TS_SYNC_BYTE we skipped earlier */
            FILE_READ(fp, (buf+num), ((TS_PKT_SIZE*MPEG2_SYS_BUF_PKTS)-num), i);
            num        += i;
            num        /= TS_PKT_SIZE;    /* number of packets in the buffer */
            i           = 0;              /* new position of this packet */
            aligned_pkt = buf;

            if (!num)
               break;   /* could not read the tail end of this packet */
         }

         pkt[i+1] = TS_PKT_read(pkt[i], &ts);
         if (!pkt[i+1]) {
            pkt[i+1] = pkt[i] + TS_PKT_SIZE - 1;   /* this packet is bad */
            continue;                              /* so skip it */
         }

         if (ts.PID == TS_PKT_NUL_ID)
            continue;   /* ignore NULL packets */
         else if (ts.adapt_field_pos) {
            /* get the ADAPTATION FIELD */
            if (!ADAPT_FIELD_read(ts.adapt_field_pos, &af))
               continue;   /* ADAPTATION FIELD is corrupt */
            tot_afs++;
         }

/*********** START TO DO SOMETHING WITH pkt[i], ts ***********/
         for (i2 = 0; (i2 < num_ids)&&(i2 < MPEG2_SYS_PIDS); i2++) {
            if (ts_pkt[i2].PID == ts.PID) {
               if (ts.continuity_counter == (cont_cnt[i2]&TS_PKT_CONT_CNT_MASK)) {
                  /* packet contains the expected data */
                  if ( (ts.adapt_field_pos)&&(ADAPT_FIELD_PCR_FLAG( (&af) )) ) {
                     if (num_PCRs[i2] > 0) {
                        rate  = TS_PKT_SIZE*8.0 * (double)(tot_pkts+i-last_pkt[i2]);
                        rate /= (double)(af.PCR - last_PCR[i2]);
                        if (rate > 0.0) {
                           avg[i2] += rate;
                           if (rate < min[i2])
                              min[i2] = rate;
                           if (rate > max[i2])
                              max[i2] = rate;
                        }
                        else
                           num_PCRs[i2]--;
                     }
                     num_PCRs[i2]++;
                     last_pkt[i2] = tot_pkts + i;
                     last_PCR[i2] = af.PCR;
                  }
GOT_EXPECTED_PACKET:
                  if (ts.adapt_field_pos)
                     ts_pkt[i2].payload++;
                  ts_pkt[i2].continuity_counter = ts.continuity_counter;
                  if (ts.payload)
                     cont_cnt[i2]++;
               }
               else if (ts.continuity_counter != ts_pkt[i2].continuity_counter) {
                  /* counter discontinuity */
                  printf("Discontinuity> PID=0x%04hX"
                         "  CCexp=0x%02hX  CCact=0x%02hX  ",
                           ts_pkt[i2].PID,
                           (cont_cnt[i2]&TS_PKT_CONT_CNT_MASK),
                           ts.continuity_counter);
                  cont_cnt[i2] = ts.continuity_counter;

                  if ( (ts.adapt_field_pos)&&(ADAPT_FIELD_PCR_FLAG( (&af) )) ) {
                     last_pkt[i2] = tot_pkts + i;
                     last_PCR[i2] = af.PCR;
                     printf("af.flag=0x%02hX\n",
                              (u16)ADAPT_FIELD_DISCONTINUITY_FLAG( (&af) ));
                  }
                  else
                     printf("\n");
                  goto GOT_EXPECTED_PACKET;
               }
               break;
            }
         }

         if (i2 == num_ids) {
            /* found a new ID */
            if (i2 < MPEG2_SYS_PIDS) {
               memcpy((ts_pkt+i2), &ts, sizeof(ts));
               cont_cnt[i2] = ts.continuity_counter;
               if (ts.payload)
                  cont_cnt[i2]++;

               if (ts.adapt_field_pos) {
                  ts_pkt[i2].payload = (const u8 *)1;
                  if (ADAPT_FIELD_PCR_FLAG( (&af) )) {
                     last_pkt[i2] = tot_pkts + i;
                     num_PCRs[i2] = 1;
                     last_PCR[i2] = af.PCR;
                  }
               }
               else
                  ts_pkt[i2].payload = (const u8 *)0;
            }
            num_ids++;
         }
/********** STOP TO DOING SOMETHING WITH pkt[i], ts **********/
      }

      tot_pkts += num;

      TIMER_curtime(&t2);
      TIMER_diff(&t1, &t2);
      TIMER_add(&tot_t, &t1);

      if (num < MPEG2_SYS_BUF_PKTS)
         break;
      else
         pkt[0] = buf;
   }

   printf("  Data Rates\n");
   printf("    PID   AFs PCRs      Min      Avg      Max\n");
   for (i = 0; (i < num_ids)&&(i < MPEG2_SYS_PIDS); i++) {
      printf("   %04hX  %4u %4d  ", ts_pkt[i].PID,
                               (u32)ts_pkt[i].payload,
                                    num_PCRs[i]);

      if (num_PCRs[i] > 1) {
         min[i] *= MPEG2_SYS_CLK_FREQ / 1.0e6;
         max[i] *= MPEG2_SYS_CLK_FREQ / 1.0e6;
         avg[i] *= (MPEG2_SYS_CLK_FREQ / 1.0e6) / (double)(num_PCRs[i] - 1);
         printf("%7.3f  %7.3f  %7.3f\n", min[i], avg[i], max[i]);
      }
      else
         printf("    N/A      N/A      N/A\n");
   }
   printf("* All rates have units of Mbits/sec.\n\n");

   printf("TS Packets = %6d        ", tot_pkts);
   printf("ADAPTATION FIELDs = %d\n", tot_afs);
   printf("Tot Time   = %8.3fmsec\n", (1.0e3*TIMER_getdbl(&tot_t)));
   printf("Avg Time   = %8.3fusec/TS_PKT\n", ((1.0e6*TIMER_getdbl(&tot_t)) / (double)tot_pkts));
   printf("Bit Rate   = %8.3fMbits/sec\n", (((double)tot_pkts*TS_PKT_SIZE*8.0) / (1.0e6*TIMER_getdbl(&tot_t))));
   if (num_ids > MPEG2_SYS_PIDS)
      printf("IDs        = %4d+\n\n", MPEG2_SYS_PIDS);
   else
      printf("IDs        = %4d\n\n", num_ids);
}
#endif

static void Display_PAT(FILE_PTR fp)
{
   u16 i, num, pid;
   
   /* load the PAT */
   sys.program_number = 0;
   MPEG_TS_load_PAT(fp);

   /* display PAT */
   printf("  Program Access Table\n");
   printf("   Program  PMT_PID\n");
   i = 0;
   while (1) {
      num = sys.get_program(sys.psi_pes[sys.PAT_idx].common.mem, i);
      if (!num)
         break;
      pid = sys.get_pid(sys.psi_pes[sys.PAT_idx].common.mem, num);
      printf("      %04hX     %04hX\n", num, pid);
      i++;
   }
   printf("\n");
}

static void Display_PMT(FILE_PTR fp, u16 prog_num)
{
   PMT_SECTION *pmt;
   u8           i, pcr;
   
   /* load the PAT */
   sys.program_number = 0;
   MPEG_TS_load_PAT(fp);
   
   /* load the PMT */
   sys.program_number = prog_num;
   MPEG_TS_load_PMT(fp);
   pmt = sys.get_PMT(sys.psi_pes[sys.PMT_idx].common.mem);
   if (!pmt) {
      printf("PMT for program (%04hX) cannot be found.\n\n", prog_num);
      return;
   }

   printf("  Program Map Table (%04hX)\n", prog_num);
   printf("    PID  Type\n");
   for (i = 0, pcr = 0; i < pmt->elem_size; i++) {
      printf("   %04hX  ", pmt->elem[i]->elementary_PID);
      if (pmt->elem[i]->elementary_PID == pmt->PCR_PID) {
         printf("PCR, ");
         pcr = 1;
      }
      Display_stream_type(pmt->elem[i]->elementary_PID,
                          pmt->elem[i]->stream_type);
   }
   if ( (i)&&(!pcr) )
      printf("   %04hX  Program Clock Reference\n\n", pmt->PCR_PID);
   else
      printf("\n");
}

static void Display_stream_type(u16 PID, u8 stream_type)
{
   if (PID == TS_PKT_PAT_ID)
      printf("Program Association Table\n");
   else if (PID == TS_PKT_CAT_ID)
      printf("Conditional Access Table\n");
   else if (PID == TS_PKT_ATSC_SI_ID)
      printf("ATSC System Info & Program Guide Data\n");
   else {
      switch (stream_type) {
      case STREAM_TYPE_MPEG1_VIDEO:
         printf("MPEG1 Video\n");
         break;
      case STREAM_TYPE_MPEG2_VIDEO:
         printf("MPEG2 Video\n");
         break;
      case STREAM_TYPE_MPEG1_AUDIO:
         printf("MPEG1 Audio\n");
         break;
      case STREAM_TYPE_MPEG2_AUDIO:
         printf("MPEG2 Audio\n");
         break;
      case STREAM_TYPE_PRIVATE:
         printf("Private\n");
         break;
      case STREAM_TYPE_PES_PRIVATE:
         printf("Private PES\n");
         break;
      case STREAM_TYPE_ISO_13522_MHEG:
         printf("ISO 13522 MHEG\n");
         break;
      case STREAM_TYPE_ISO_13818_1_DSMCC:
         printf("ISO 13818 1 DSMCC\n");
         break;
      case STREAM_TYPE_ITUT_H2221:
         printf("ITUT H2221\n");
         break;
      case STREAM_TYPE_ISO_13818_6_A:
         printf("ISO 13818 6 A\n");
         break;
      case STREAM_TYPE_ISO_13818_6_B:
         printf("ISO 13818 6 B\n");
         break;
      case STREAM_TYPE_ISO_13818_6_C:
         printf("ISO 13818 6 C\n");
         break;
      case STREAM_TYPE_ISO_13818_6_D:
         printf("ISO 13818 6 D\n");
         break;
      case STREAM_TYPE_ISO_13818_1_AUX:
         printf("ISO 13818 1 AUX\n");
         break;
      case STREAM_TYPE_ATSC_AC3_AUDIO:
         printf("ATSC AC3 Audio\n");
         break;
      default:
         if ( (0x80 <= stream_type)&&(stream_type <= 0xAF) )
            printf("ATSC Private Stream\n");
         else
            printf("Unknown\n");
      }
   }
}

void display_time(double sec)
{
   int hrs = (int)(sec / 3600.0), 
       min = (int)(sec /   60.0);

   if (hrs > 0) {
      printf("%dh%02dm%06.3lfs",
             hrs, (min - (hrs*60)),
             (sec - (double)(min*60)));
   }
   else if (min > 0) {
      printf("%dm%06.3lfs",
             min, (sec - (double)(min*60)));
   }
   else
      printf("%.3lfs", sec);
}

#ifdef _RECORD_STATS_
static void Display_stats()
{
   /* display stream length */
   if (sys.data.v_pos > 0) {
      printf("\n  Video Time   : ");
      display_time((double)sys.data.v_pos/(double)MPEG2_SYS_CLK_FREQ);
   }
   if (sys.data.a_pos > 0) {
      printf("\n  Audio Time   : ");
      display_time((double)sys.data.a_pos/(double)MPEG2_SYS_CLK_FREQ);
   }
   
#if (MPEG2_SYS_SLOW_DOWN > 1)
   sys.decode_time /= MPEG2_SYS_SLOW_DOWN;
#endif
   if (sys.decode_time > 0.0) {
      double rate = (double)(8*sys.stream_bytes) / sys.decode_time;

      /* display decoding time */
      printf("\n  Decoding Time: ");
      display_time(sys.decode_time);
#if (MPEG2_SYS_SLOW_DOWN > 1)
      printf(" (x%d)", MPEG2_SYS_SLOW_DOWN);
#endif

      /* display bit rate */
      printf("\n  Bit Rate     : ");
      if (rate > 1.0e9)
         printf("%.3lfGbps", (rate / 1.0e9));
      else if (rate > 1.0e6)
         printf("%.3lfMbps", (rate / 1.0e6));
      else if (rate > 1.0e3)
         printf("%.3lfKbps", (rate / 1.0e3));
      else
         printf("%.3lfbps", rate);

      if (sys.pkt_cnt) {
         /* display packet count */
         printf("\n  Packet Count : %u", sys.pkt_cnt);

         /* display packet processing time */
         printf("\n  Packet Time  : ");
         rate = sys.decode_time / (double)sys.pkt_cnt;
         if (rate > 1.0)
            printf("%.3lfsec/pkt", rate);
         else if (rate > 1.0e-3)
            printf("%.3lfmsec/pkt", (rate * 1.0e3));
         else
            printf("%.3lfnsec/pkt", (rate * 1.0e6));
      }
      printf("\n\n");
   }
}
#endif

#undef NUM_PIDS

